/**
 * 
 */
package sim.geometry;

import java.util.concurrent.atomic.AtomicLong;

import sim.exception.SRuntimeException;
import sim.graphics.SPrimitive;
import sim.math.SVector3d;
import sim.math.SVectorUV;
import sim.readwrite.SAbstractReadableWriteable;
import sim.util.SInitializationException;

/**
 * La classe abstraite <b>SAbstractGeometry</b> reprsentant une gomtrie gnrale sans dfinition spatiale. 
 *  
 * @author Simon Vzina
 * @since 2014-12-30
 * @version 2020-02-11
 */
public abstract class SAbstractGeometry extends SAbstractReadableWriteable implements SGeometry {

  //--------------
  // CONSTANTES //
  //--------------
  
  //Numro d'identification des types de gomtrie 
  public static final int PLANE_CODE = 1;
  public static final int DISK_CODE = 2;
  public static final int SPHERE_CODE = 3;
  public static final int TUBE_CODE = 4;
	public static final int CYLINDER_CODE = 5;
	public static final int CONE_CODE = 6;
	public static final int TRIANGLE_CODE = 7;
	public static final int BTRIANGLE_CODE = 8;
	public static final int TRANSFORMABLE_CODE = 9;
	public static final int CUBE_CODE = 10;
	public static final int SPHERICAL_CAP_CODE = 11;
	public static final int LENS_CODE = 12;
	public static final int TORUS_CODE = 13;
	
	//-------------
	// VARIABLES //
	//-------------
	
	/**
	 * La variable <b>atomic_counteur_ID</b> correspond  un compteur THREAD PROOF  numro d'identification des gomtries. 
	 */
	private static final AtomicLong atomic_counter_ID = new AtomicLong();
  
	/**
   * La variable <b>id</b> correspond au numro d'identification unique de la gomtrie.
   */
	private long id;						           
	
	/**
	 * La variable <b>primitive_parent</b> correspond  la primitive qui est le propritaire de la gomtrie (son parent).
	 */
	private SPrimitive primitive_parent;	 
	
	//----------------
	// CONSTRUCTEUR //
	//----------------
	
	/**
	 * Constructeur d'une gomtrie sans parent primitive.
	 */
	public SAbstractGeometry()
	{
		this(null);
	}
	
	/**
	 * Constructeur par dfaut d'une gomtrie ayant une primitive comme parent.
	 * 
	 * @param parent La primitive parent de la gomtrie.
	 */
	public SAbstractGeometry(SPrimitive parent)
	{
		// Mettre  jour le compter et obtenir le nouveau ID (THREAD PROOF)
	  id = atomic_counter_ID.incrementAndGet();
	  
		primitive_parent = parent;
	}
	
	@Override
	public long getID()
	{ 
	  return id;
	}
	
	@Override
	public int hashCode()
	{
	  return (int) id;
	}
	
	@Override
  public SPrimitive getPrimitiveParent()throws SRuntimeException
  { 
    if(primitive_parent != null)
      return primitive_parent;
    else
      throw new SRuntimeException("Erreur SAbstractGeometry 001 : La gomtrie ne possde pas de primitive comme parent.");
    
  }
  
  @Override
  public void setPrimitiveParent(SPrimitive parent)throws SRuntimeException
  {
    if(primitive_parent != null)
      throw new SRuntimeException("Erreur SAbstractGeometry 002 : La gomtrie ne peut pas se faire affecter un nouveau parent, car cette valeur a t pralablement dtermine.");
    else
      primitive_parent = parent;
  }
  
	@Override
	public boolean isTransparent()
	{
	  if(primitive_parent != null)
	    if(primitive_parent.getMaterial() != null)
	      return primitive_parent.getMaterial().isTransparent();
	    else
	      return false;
	  else
	    return false;
  } 
		
	/**
	 * Mthode pour dterminer la normale  la surface de la gomtrie intersecte par le rayon.
	 * Cette mthode doit dterminer <u>l'orientation extrieure</u>  la gomtrie.
	 * 
	 * @param ray Le rayon ralisant l'intersection avec la gomtrie.
	 * @param intersection_t Le temps requis au rayon pour raliser l'intersection. 
	 * @return La normale  la surface <b>normalise</b> o est ralise l'intersection.
	 */
	abstract protected SVector3d evaluateIntersectionNormal(SRay ray, double intersection_t);
	
	/**
	 * Mthode pour dterminer les coordonnes uv de la gomtrie  l'endroit o il y a intersectin avec un rayon.
	 * Avant d'utiliser cette mthode, il faut pralablement valuer le temps pour raliser l'intersection.
   * Cette mthode prendra pour acquis que ces tests ont t raliss adquatement.
   * 
	 * @param ray Le rayon ralisant l'intersection avec la gomtrie.
	 * @param intersection_t Le temps requis au rayon pour raliser l'intersection. 
	 * @return La coordonne uv de la surface o est ralise l'intersection.
	 */
	abstract protected SVectorUV evaluateIntersectionUV(SRay ray, double intersection_t);
	
	@Override
	public boolean equals(Object other)
	{
		if (other == null) 					//Test du null
		  return false;
	    
	  if (other == this)					//Test de la mme rfrence 
	    return true;
	    
	  if (!(other instanceof SGeometry))	//Test d'un type diffrent
	    return false;
	  
	  //------------------------------------------------------------------------------------
	  // REMARQUE : Puisque chaque gomtrie aura sa propre ID, je ne peux pas utiliser
	  //            ce paramtre dans le equals(...), car autrement, ce sera toujours faux.
	  //------------------------------------------------------------------------------------
	  
	  /*
	  //Vrification du numro d'identification, car il doit tre unique pour chaque gomtrie construite
	  SAbstractGeometry o = (SAbstractGeometry)other;
	    
	  if(id != o.getID())
	    return false;
	   */
	  
	  return true;
	}
			
	@Override
  protected void readingInitialization() throws SInitializationException
  {
   
  }
		
	@Override
  public String toString()
  {
    return "SAbstractGeometry [name=" + getReadableName() + ", id=" + id + "]";
  }
  
}//fin de la classe abstraite SAbstractGeometry
