/**
 * 
 */
package sim.geometry;

import sim.exception.SRuntimeException;
import sim.math.SImpossibleNormalizationException;
import sim.math.SMath;
import sim.math.SVector3d;
import sim.util.SLog;

/**
 * <p>
 * La classe <b>SGeometricUtil</b> reprsente une classe utilitaire permettant d'analyser des gomtries.
 * </p>
 * 
 * <p>
 * On y retrouve des mthodes comme :
 * <ul>- valuation si un vecteur <i>v</i> est  l'intrieur, sur ou  l'extrieur d'un <u>primtre</u>.</ul>
 * <ul>- valuation si un vecteur <i>v</i> est  l'intrieur, sur ou  l'extrieur d'une <u>surface</u>.</ul>
 * </p>
 * 
 * @author Simon Vzina
 * @since 2015-11-23
 * @version 2022-01-12
 */
public class SGeometricUtil {

  /**
   * <p>
   * Mthode pour dterminer si une coordonne (x,y) se retrouve sur le primtre d'un rectangle
   * centr  la coordonne (x_R, y_R) avec des dimensions w par h.
   * </p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si (x,y) est  l'intrieur du primtre, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si (x,y) est sur le primtre, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si (x,y) est  l'extrieur du primtre, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param x_R La position centrale du rectangle selon l'axe x (largeur).
   * @param y_R La position centrale du rectangle selon l'axe y (hauteur).
   * @param w La largeur du rectangle. Cette valeur doit tre <b>positive</b>.
   * @param h La hauteur du rectangle. Cette valeur doit tre <b>positive</b>.
   * @param x La coordonne x  analyser.
   * @param y La coordonne y  analyser.
   * @return <b>-1</b> si (x,y) est  l'intrieur, <b>0</b> si (x,y) est sur le primtre et <b>1</b> si (x,y) est  l'extrieur du primtre.
   * @throws SRuntimeException Si la largeur w ou la hauteur h est ngative.
   */
  public static int isOnRectanglePerimeter(double x_R, double y_R, double w, double h, double x, double y) throws SRuntimeException
  {
    if(w < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 001 : La valeur de w = " + w + " est ngative.");
    
    if(h < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 002 : La valeur de h = " + h + " est ngative.");
      
    // valuer les distances maximales du rectangle  partir de son centre
    double half_w_plus = SMath.ONE_PLUS_EPSILON*w/2;
    double half_w_minus = SMath.ONE_MINUS_EPSILON*w/2;
    
    double half_h_plus = SMath.ONE_PLUS_EPSILON*h/2;
    double half_h_minus = SMath.ONE_MINUS_EPSILON*h/2;
    
    // Distance selon l'axe x entre le point et le centre du rectangle
    double dx = Math.abs(x - x_R);
    
    // Vrifier si la distance est  l'extrieur selon l'axe x
    if(dx > half_w_plus)
      return 1;
    
    // Distance selon l'axe y entre le point et le centre du rectangle
    double dy = Math.abs(y - y_R);
    
    // Vrifier si la distance est  l'extrieur selon l'axe y
    if(dy > half_h_plus)
      return 1;
    
    // Vrifier si la distance est  l'intrieur selon x et y
    if(dx < half_w_minus && dy < half_h_minus)
      return -1;
    
    // Le dernier cas reste d'tre sur le primtre
    return 0;
  }
  
  /**
   * <p>
   * Mthode pour dterminer si une coordonne (x,y) se retrouve sur le primtre d'un cercle
   * centr  la coordonne (x_C, y_C) avec un rayon R.
   * </p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si (x,y) est  l'intrieur du primtre, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si (x,y) est sur le primtre, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si (x,y) est  l'extrieur du primtre, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param x_C La position centrale du cercle selon l'axe x (largeur).
   * @param y_C La position centrale du cercle selon l'axe y (hauteur).
   * @param R Le rayon du cercle. Cette valeur doit tre <b>positive</b>.
   * @param x La coordonne x  analyser.
   * @param y La coordonne y  analyser.
   * @return <b>-1</b> si (x,y) est  l'intrieur, <b>0</b> si (x,y) est sur le primtre et <b>1</b> si (x,y) est  l'extrieur du primtre.
   * @throws SRuntimeException Si le rayon R est ngatif.
   */
  public static int isOnCerclePerimeter(double x_C, double y_C, double R, double x, double y) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 003 : La valeur de R = " + R + " est ngative.");
    
    // Puisque la mthode de l'ellipse utilise le diamtre, il faudra multiplier le rayon par deux (D = 2R).
    return isOnEllipsePerimeter(x_C, y_C, 2*R, 2*R, x, y);
  }
  
  /**
   * <p>
   * Mthode pour dterminer si une coordonne (x,y) se retrouve  l'intrieur d'une ellipse
   * centr  la coordonne (x_E, y_E) avec des dimensions w par h.
   * </p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si (x,y) est  l'intrieur du primtre, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si (x,y) est sur le primtre, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si (x,y) est  l'extrieur du primtre, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param x_E La position centrale de l'ellispse selon l'axe x (largeur).
   * @param y_E La position centrale de l'ellispse selon l'axe y (hauteur).
   * @param w Longueur du grand diamtre (le double du demi-grand axe) selon l'axe x. Cette valeur doit tre <b>positive</b>.
   * @param h Longueur du petit diamtre (le double du demi-petit axe) selon l'axe y. Cette valeur doit tre <b>positive</b>.
   * @param x La coordonne x  analyser.
   * @param y La coordonne y  analyser.
   * @return <b>-1</b> si (x,y) est  l'intrieur, <b>0</b> si (x,y) est sur le primtre et <b>1</b> si (x,y) est  l'extrieur du primtre.
   * @throws IllegalArgumentException Si la longeur du grand diamtre w ou la longueur du petit diamtre h est ngative.
   */
  public static int isOnEllipsePerimeter(double x_E, double y_E, double w, double h, double x, double y) throws IllegalArgumentException
  {
    if(w < 0)
      throw new IllegalArgumentException("Erreur SGeometricUtil 004 : La valeur de w = " + w + " est ngative.");
    
    if(h < 0)
      throw new IllegalArgumentException("Erreur SGeometricUtil 005 : La valeur de h = " + h + " est ngative.");
     
    // valuer l'ellipse :
    // Si gale  1     => sur l'ellipse
    // Si infrieur  1 =>  l'intrieur de l'ellipse
    // Si suprieur  1 =>  l'extrieur de l'ellipse
    double value = (x - x_E)*(x - x_E) / (w*w/4) + (y - y_E)*(y - y_E) / (h*h/4);
    
    // P.S. Dans la littrature, w et h correspond  la moitier des axes (au rayon). 
    //      C'est pourquoi il faut diviser par 4 dans la formule,
    //      car les paramtres de la mthode correspond au diamtre et non au rayon (r = D/2).
    return compareDistanceWithReference(value, 1.0);
  }
  
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface de deux plans parallles.</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur des deux surfaces, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur l'une des deux surfaces, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur des deux surfaces, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param r_plane1 La position du plan 1.
   * @param r_plane2 La position du plan 2.
   * @param axis_1to2 Un axe passant du plan 1 au plan 2 donnant ainsi l'alignement des plans (ceux-ci tant perpendiculaire  cet axe). L'axe doit tre <b>normalis</b>.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   */
  public static int isOnTwoParallelsPlanesSurface(SVector3d r_plane1, SVector3d r_plane2, SVector3d axis_1to2, SVector3d v)
  {
    // S'assurer de la normalisation de l'axe
    axis_1to2 = axis_1to2.normalize();
    
    // Vrifier si le vecteur v est  l'extrieur du plan 1
    SVector3d P1_to_v = v.substract(r_plane1);
    double distance1 = P1_to_v.dot(axis_1to2);
    
    //   l'extrieur du plan 1
    if(distance1 < SMath.NEGATIVE_EPSILON)
      return 1;
    else
      // Trop prs du plan 1
      if(distance1 < SMath.EPSILON)
        return 0;
    
    //Vrifier si le vecteur v est  l'extrieur du plan 2
    SVector3d P2_to_v = v.substract(r_plane2);
    double distance2 = P2_to_v.dot(axis_1to2);
    
    //  l'extrieur du plan 2
    if(distance2 > SMath.EPSILON)
      return 1;
    else
      // Trop prs du plan 2
      if(distance2 > SMath.NEGATIVE_EPSILON)
        return 0;
    
    // Aprs le test sur les deux plans, le vecteur v se retrouve entre les deux plans.
    return -1;
  }
  
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface d'une sphre.</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur de la surface, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur la surface, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur de la surface, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param r_s La position de la sphre.
   * @param R Le rayon de la sphre. Cette valuer doit tre <b>positive</b>.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   * @throws SRuntimeException Si le rayon R est ngatif.
   */
  public static int isOnSphereSurface(SVector3d r_s, double R, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 006 : La valeur de R = " + R + " est ngative.");
    
    // valuer la distance entre le centre de la sphre et le vecteur v
    double distance = v.substract(r_s).modulus();  
    
    return compareDistanceWithReference(distance, R);
  }
  
  /**
   * Mthode pour obtenir la position d'un point sur une sphre  partir de deux angles.
   * 
   * @param r_S La position de la sphre.
   * @param R Le rayon de la sphre.
   * @param phi L'angle dans le plan xy (entre 0 et 2Pi).
   * @param theta L'angle selon l'axe z (entre 0 et Pi).
   * @return La position sur la sphre.
   */
  public static SVector3d sphericalCoordinate(SVector3d r_S, double R, double phi, double theta)
  {
    double x = R*Math.cos(phi)*Math.sin(theta);
    double y = R*Math.sin(phi)*Math.sin(theta);
    double z = R*Math.cos(theta);
    
    SVector3d r = new SVector3d(x, y, z);
    
    return r_S.add(r);
  }
  
  /**
   * Mthode pour valuer la normale  la <u>surface extrieure</u> d'une sphre  partir d'un point <b>sur la sphre</b>. 
   * La normale calcule sera normalise.
   * 
   * @param rS La position de la sphre.
   * @param R Le rayon de la sphre. Cette valuer doit tre <b>positive</b>.
   * @param v Le vecteur position o l'on veut valuer la normale  la surface.
   * @return La normale  la surface extrieure d'une sphre.
   * @throws SRuntimeException Si le rayon R est ngatif.
   * @throws SRuntimeException Si le vecteur v n'est pas sur la surface.
   */
  public static SVector3d outsideSphereNormal(SVector3d rS, double R, SVector3d v) throws IllegalArgumentException
  {
    if(R < 0)
      throw new IllegalArgumentException("Erreur SGeometricUtil 006 : La valeur de R = " + R + " est ngative.");
    
    // valuer le vecteur distance entre le centre de la sphre et le vecteur v.
    SVector3d rS_to_v = v.substract(rS);
    
    // Vrifier que le vecteur v est sur la surface.
    if(compareDistanceWithReference(rS_to_v.modulus(), R) != 0)
      throw new IllegalArgumentException("Erreur SGeometricUtil 007 : Le vecteur v = " + v + " n'est pas sur la sphre (rS = " + rS + ", R = " + R + ").");
    
    return rS_to_v.normalize();
  }
   
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface d'un tube infini.</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur de la surface, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur la surface, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur de la surface, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param rT Un point sur l'axe du tube infini.
   * @param axis L'orientation de l'axe du tube infini. L'axe doit tre <b>normalis</b>.
   * @param R Le rayon du tube infini. Cette valuer doit tre <b>positive</b>.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   * @throws SRuntimeException Si le rayon R est ngatif.
   */
  public static int isOnInfiniteTubeSurface(SVector3d rT, SVector3d axis, double R, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 007 : La valeur de R = " + R + " est ngative.");
    
    // S'assurer de la normalisation de l'axe
    axis = axis.normalize();
    
    // valuons la distance entre le vecteur v et l'axe du tube
    double distance = axis.cross( v.substract(rT) ).cross(axis).modulus();
    
    return compareDistanceWithReference(distance, R, SMath.ONE_PLUS_1000EPSILON, SMath.ONE_MINUS_1000EPSILON);
  }
  
  /**
   * Mthode pour valuer la normale  la <u>surface extrieure</u> d'un tube  partir d'un point <b>sur le tube</b>. 
   * La normale calcule sera normalise.
   *
   * @param rT Un point sur l'axe du tube infini.
   * @param axis L'orientation de l'axe du tube infini. L'axe doit tre <b>normalis</b>.
   * @param R Le rayon du tube infini. Cette valuer doit tre <b>positive</b>.
   * @param v Le vecteur position  analyser.
   * @return La normale  la surface extrieure d'un tube.
   * @throws SRuntimeException Si le rayon R est ngatif.
   * @throws SRuntimeException Si le vecteur v n'est pas sur la surface.
   */
  public static SVector3d outsideInfiniteTubeNormal(SVector3d rT, SVector3d axis, double R, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 007 : La valeur de R = " + R + " est ngative.");
    
    // S'assurer de la normalisation de l'axe
    axis = axis.normalize();
    
    // valuer le vecteur distance la plus courte entre l'axe du tube et le  vecteur v.
    SVector3d axis_to_v = axis.cross( v.substract(rT) ).cross(axis);
    
    // Vrifier que le vecteur v est sur la surface.
    if(compareDistanceWithReference(axis_to_v.modulus(), R, SMath.ONE_PLUS_1000EPSILON, SMath.ONE_MINUS_1000EPSILON) != 0)
      throw new SRuntimeException("Erreur SGeometricUtil 007 : Le vecteur v = " + v + " n'est pas sur le tube (rT = " + rT + ", axis = " + axis + ", R = " + R + ") et la distance v-axis est " + axis_to_v.modulus() + ".");
    
    // Retourner l'orientation extrieur de la normale du tube normalise.
    return axis_to_v.normalize();
  }
  
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface de deux cnes infinis (dont les pointes se touchent comme un sablier).</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur de la surface, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur la surface, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur de la surface, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param r_c Une position sur l'axe central des deux cnes o le rayon <i>R</i> a t dfini.
   * @param axis L'axe des deux cnes dans la direction localisant la pointe du cne  partir de la position <i>r_c</i>. L'axe doit tre <b>normalis</b>.
   * @param R Le rayon du cne  la position <i>r_c</i>. Cette valuer doit tre <b>positive</b>.
   * @param H La hauteur du cne tant dfinie comme la distance entre la position <i>r_c</i> et la pointe des cnes. Cette valuer doit tre <b>positif</b>.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   * @throws SRuntimeException Si le rayon R ou la hauteur H du cne est ngatif.
   */
  public static int inOnTwoInfinitesConesSurface(SVector3d r_c, SVector3d axis, double R, double H, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 008 : La valeur de R = " + R + " est ngative.");
    
    if(H < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 009 : La valeur de H = " + H + " est ngative.");
    
    // S'assurer de la normalisation de l'axe
    axis = axis.normalize();
    
    // valuons la distance entre le vecteur v et l'axe du cne
    double distance = axis.cross( v.substract(r_c) ).cross(axis).modulus();
    
    // valuons h tant la distance entre le point de rfrence sur l'axe du cne et la projection de v sur l'axe du cne
    double h = v.substract(r_c).dot(axis);  // distance entre le point de 
    
    // valuer le rayon du cne  l'endroit de la projection du vecteur v sur l'axe du cne
    // Cette valeur sera gale  (1 - h/H)*R et devra tre positif (si la projection h est plus grande que H)
    double R_var = Math.abs((1 - h/H)*R);
    
    //return compareDistanceWithReference(distance, R_var); // Ancien seuil pas assez relax
    return compareDistanceWithReference(distance, R_var, SMath.ONE_PLUS_1000EPSILON, SMath.ONE_MINUS_1000EPSILON);
  }
  
  /**
   * 
   * @param rC Une position sur l'axe central des deux cnes o le rayon <i>R</i> a t dfini.
   * @param axis L'axe des deux cnes dans la direction localisant la pointe du cne  partir de la position <i>r_c</i>. L'axe doit tre <b>normalis</b>.
   * @param R Le rayon du cne  la position <i>rC</i>. Cette valuer doit tre <b>positive</b>.
   * @param H La hauteur du cne tant dfinie comme la distance entre la position <i>rC</i> et la pointe des cnes. Cette valuer doit tre <b>positif</b>.
   * @param v Le vecteur position  analyser.
   * @return
   * @throws SRuntimeException Si le rayon R ou la hauteur H du cne est ngatif.
   * @throws SRuntimeException Si le vecteur v n'est pas sur la surface.
   */
  public static SVector3d outsideTwoInfinitesConesNormal(SVector3d rC, SVector3d axis, double R, double H, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 011 : La valeur de R = " + R + " est ngative.");
    
    if(H < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 012 : La valeur de H = " + H + " est ngative.");
   
    // S'assurer que le vecteur v est sur la surface.
    if(inOnTwoInfinitesConesSurface(rC, axis, R, H, v) != 0)
      throw new SRuntimeException("Erreur SGeometricUtil 013 : Le vecteur v = " + v + " n'est pas sur le double cne (rC = " + rC + ", axis = " + axis  + ", R = " + R + ", H = " + H + ").");
    
    try{
      
      // S'assurer de la normalisation de l'axe
      axis = axis.normalize();
      
      SVector3d top = rC.add(axis.multiply(H));
      
      SVector3d edge = top.substract(v).normalize();
      
      SVector3d ext_normal = edge.cross(axis).cross(edge);
      
      return ext_normal.normalize();
      
    }catch(SImpossibleNormalizationException e){
      //Si l'intersection est trop prs du point P2, alors un trouvera une rponse  a !!!
      SLog.logWriteLine("Message SGeometricUtil : Normale value  la pointe du cne.");
      return axis;
    }
     
  }
  
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface d'un cylindre.</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur de la surface, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur la surface, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur de la surface, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param r_1 Le 1ier point dlimitant l'axe du cylindre.
   * @param r_2 Le 2ime point dlimitant l'axe du cylindre.
   * @param R Le rayon du cylindre. Cette valuer doit tre <b>positive</b>.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   * @throws SRuntimeException Si le rayon R est ngatif.
   */
  public static int isOnCylinderSurface(SVector3d r_1, SVector3d r_2, double R, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 010 : La valeur de R = " + R + " est ngative.");
    
    // Dterminer l'axe du cylindre
    SVector3d axis = r_2.substract(r_1).normalize();
        
    // Dterminer les codes des extrmit et du tube
    int code_extremity = isOnTwoParallelsPlanesSurface(r_1, r_2, axis, v);
    int code_tube = isOnInfiniteTubeSurface(r_1, axis, R, v);
    
    // Cas #1 :  l'extrieur des extrmits du cylindre ou du tube
    if(code_extremity > 0 || code_tube > 0)
      return 1;
    
    // Case #2 :  l'intrieur des extrmits du cylindre et du tube
    if(code_extremity < 0 && code_tube < 0)
      return -1;
     
    // Case #3 : Sur la surface des disques du cylindre ou sur le tube
    return 0;
  }
  
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface d'un cne.</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur de la surface, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur la surface, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur de la surface, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param r_c Une position sur l'axe central de la base du cne o le rayon <i>R</i> a t dfini.
   * @param axis L'axe du cne dans la direction localisant la pointe du cne  partir de la position <i>r_c</i>. L'axe doit tre <b>normalis</b>.
   * @param R Le rayon du cne  la position <i>r_c</i>. Cette valuer doit tre <b>positive</b>.
   * @param H La hauteur du cne tant dfinie comme la distance entre la position <i>r_c</i> et la pointe des cnes. Cette valuer doit tre <b>positif</b>.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   * @throws SRuntimeException Si le rayon R ou la hauteur H du cne est ngatif.
   */
  public static int isOnConeSurface(SVector3d r_c, SVector3d axis, double R, double H, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 011 : La valeur de R = " + R + " est ngative.");
    
    if(H < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 012 : La valeur de H = " + H + " est ngative.");
    
    // S'assurer de la normalisation de l'axe
    axis = axis.normalize();
    
    // Dterminer les codes des extrmit et du cne
    int code_extremity = isOnTwoParallelsPlanesSurface(r_c, r_c.add(axis.multiply(H)), axis, v);
    int code_cone = inOnTwoInfinitesConesSurface(r_c, axis, R, H, v);
    
    // Cas #1 :  l'extrieur des extrmits du cone ou  l'extrieur de la forme cnique
    if(code_extremity > 0 || code_cone > 0)
      return 1;
    
    // Case #2 :  l'intrieur des extrmits du cone et  l'intrieur de la forme cnique
    if(code_extremity < 0 && code_cone < 0)
      return -1;
     
    // Case #3 : Sur la surface de la base du cne ou sur la surface cnique
    return 0;
  }
    
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface d'un cube.</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur de la surface, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur la surface, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur de la surface, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param rC La position centrale du cube.
   * @param size La taille du cube.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   * @throws SRuntimeException Si la taille est ngatif.
   */
  public static int isOnAlignedCubeSurface(SVector3d rC, double size, SVector3d v)
  {
    if(size < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 013 : La taille du cube gale  " + size + " est ngative.");
    
    // valuer la position du vecteur v par rapport au centre du cube.
    SVector3d p = v.substract(rC);
        
    // Modifier les conditions de validation afin qu'un point sur le cube ne soit pas  l'intrieur
    double minus_positive_value = SMath.ONE_MINUS_EPSILON * size/2.0;
    double minus_negative_value = SMath.ONE_MINUS_EPSILON * -size/2.0;
     
    // Cas #1 : Vrifier si p est  l'intrieur de toutes les conditions.
    if(   p.getX() < minus_positive_value && p.getX() > minus_negative_value 
       && p.getY() < minus_positive_value && p.getY() > minus_negative_value
       && p.getZ() < minus_positive_value && p.getZ() > minus_negative_value)
      return -1;
    
    // Modifier les conditions de validation afin qu'un point sur le cube ne soit pas  l'extrieur
    double plus_positive_value = SMath.ONE_PLUS_EPSILON * size/2.0;
    double plus_negative_value = SMath.ONE_PLUS_EPSILON * -size/2.0;
    
    // Cas #2 : Vrifier si p est  l'extrieur d'au moins une condition.
    if(   p.getX() > plus_positive_value || p.getX() < plus_negative_value 
       || p.getY() > plus_positive_value || p.getY() < plus_negative_value
       || p.getZ() > plus_positive_value || p.getZ() < plus_negative_value)
      return 1;
    
    // Cas #3 : On est absolument sur la surface.
    return 0;  
  }
   
  /**
   * <p>Mthode permettant d'valuer si un vecteur position v est situ sur la surface d'un tore.</p>
   * <p>
   * Le rsultat de cette mthode sera interprtable de la faon suivante :
   * <ul>- Si v est  l'intrieur de la surface, la mthode retournera <b>-1</b>.</ul>
   * <ul>- Si v est sur la surface, la mthode retournera <b>0</b>.</ul>
   * <ul>- Si v est  l'extrieur de la surface, la mthode retournera <b>1</b>.</ul>
   * </p>
   * 
   * @param rT La position centrale du tore.
   * @param n La normale au plan du tore.
   * @param R Le rayon du primtre du tore ce qui reprsente la hauteur du cylindre formant le tore.
   * @param r Le rayon de la partie cylindrique du tore.
   * @param v Le vecteur position  analyser.
   * @return <b>-1</b> si le vecteur est  l'intrieur, <b>0</b> si le vecteur est sur la surface et <b>1</b> si le vecteur est  l'extrieur de la surface.
   * @throws SRuntimeException Si R ou r est ngatif.
   */
  public static int isOnTorusSurface(SVector3d rT, SVector3d n, double R, double r, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 013 : La valeur de R = " + R + " est ngative.");
    
    if(r < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 014 : La valeur de r = " + r + " est ngative.");
    
    // Vecteur dplacement du centre du tore au point v
    SVector3d rT_to_v = v.substract(rT);
    
    // Axe le long du cylindre du tore par rapport au point d'intersection
    SVector3d cylinder_axis = rT_to_v.cross(n);
    
    // Axe radial partant du centre du tore vers le centre du cylindre du tore 
    SVector3d radius_axis = n.cross(cylinder_axis).normalize();
    
    // Position centrale du beigne par rapport au point d'intersection
    SVector3d rC = rT.add(radius_axis.multiply(R));
    
    // valuer la distance entre v et l'axe du beigne
    double distance = v.substract(rC).modulus();
    
    // Pour des raisons numriques, on doit utiliser un facteur multiplicatif plus grand.
    return compareDistanceWithReference(distance, r, SMath.ONE_PLUS_1000EPSILON, SMath.ONE_MINUS_1000EPSILON);
  }
  
  
  /**
   * Mthode pour valuer la normale  la <u>surface extrieure</u> d'un tore  partir d'un point <b>sur le tore</b>. 
   * La normale calcule sera normalise.
   * 
   * @param rT La position centrale du tore.
   * @param n La normale au plan du tore.
   * @param R Le rayon du primtre du tore ce qui reprsente la hauteur du cylindre formant le tore.
   * @param r Le rayon de la partie cylindrique du tore.
   * @param v Le vecteur position  analyser.
   * @return La normale  la surface du tore  l'endroit o est situ le vecteur v sur la surface du tore.
   * @throws SRuntimeException Si R ou r est ngatif.
   * @throws SRuntimeException Si le vecteur v n'est pas sur la surface.
   */
  public static SVector3d outsideTorusNormal(SVector3d rT, SVector3d n, double R, double r, SVector3d v) throws SRuntimeException
  {
    if(R < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 013 : La valeur de R = " + R + " est ngative.");
    
    if(r < 0)
      throw new SRuntimeException("Erreur SGeometricUtil 014 : La valeur de r = " + r + " est ngative.");
    
    SVector3d rT_to_v = v.substract(rT);
    
    // Axe le long du cylindre du tore par rapport au point d'intersection
    SVector3d cylinder_axis = rT_to_v.cross(n);
    
    // Axe radial partant du centre du tore vers le centre du cylindre du tore 
    SVector3d radius_axis = n.cross(cylinder_axis).normalize();
    
    // Position centrale du beigne par rapport au point d'intersection
    SVector3d rC = rT.add(radius_axis.multiply(R));
    
    // Vecteur distance entre la position centrale du beigne et le vecteur v
    SVector3d rC_to_v = v.substract(rC);
    
    // Vrifier que le vecteur v est sur la surface.
    // Pour des raisons numriques, on doit utiliser un facteur multiplicatif plus grand.
    if(compareDistanceWithReference(rC_to_v.modulus(), r, SMath.ONE_PLUS_1000EPSILON, SMath.ONE_MINUS_1000EPSILON) != 0)
      throw new SRuntimeException("Erreur SGeometricUtil 007 : Le vecteur v = " + v + " n'est pas sur le tore (rT = " + rT + ", n = " + n  + ", R = " + R + ", r = " + r + "). La distance entre v et l'axe rotatif du tore est d = " + rC_to_v.modulus() + ".");
    
    // Retourner la normale  la surface normalise.
    return rC_to_v.normalize();
  }
  
  //----------------------
  // MTHODE UTILITAIRE //
  //----------------------
  
  /**
   * <p>Mthode qui compare une distance avec une valeur de rfrence.</p>
   * <p> 
   * Voici les rsultats admissibles de la comparaison:
   * <ul>- Si la distance est infrieure de la rfrence, la mthode retournera -1 (intrieur  la rfrence).</ul>
   * <ul>- Si la distance est gale  la rfrence ( un EPSILON prs), la mthode retournera 0 (sur la rfrence).</ul>
   * <ul>- Si la distance est suprieure  la rfrence, la mthode retournera 1 (extrieur  la rfrence).</ul>
   * </p>
   * @param distance La distance  comparer.
   * @param reference La rfrence.
   * @return <b>-1</b> si la distance est infrieure, <b>0</b> si la distance est gale et <b>1</b> si la distance est suprieure  la rfrence.
   */ 
  private static int compareDistanceWithReference(double distance, double reference)
  {
    return compareDistanceWithReference(distance, reference, SMath.ONE_PLUS_EPSILON, SMath.ONE_MINUS_EPSILON);
  }
  
  /**
   * <p>Mthode qui compare une distance avec une valeur de rfrence.</p>
   * <p> 
   * Voici les rsultats admissibles de la comparaison:
   * <ul>- Si la distance est infrieure de la rfrence, la mthode retournera -1 (intrieur  la rfrence).</ul>
   * <ul>- Si la distance est gale  la rfrence ( un EPSILON prs), la mthode retournera 0 (sur la rfrence).</ul>
   * <ul>- Si la distance est suprieure  la rfrence, la mthode retournera 1 (extrieur  la rfrence).</ul>
   * </p>
   * @param distance La distance  comparer.
   * @param reference La rfrence.
   * @param up_multiplicator Le multiplicateur suprieur  appliquer  la rfrence pour dfinir la borne suprieure de l'galit. Ce multiplicateur doit tre prs de 1, mais lgrement suprieur (ex : 1.000001).
   * @param down_multiplicator Le multiplicateur infrieur  appliquer  la rfrence pour dfinir la borne infrieur de l'galit. Ce multiplicateur doit tre prs de 1, mais lgrement infrieur (ex : 0.99999).
   * @return <b>-1</b> si la distance est infrieure, <b>0</b> si la distance est gale et <b>1</b> si la distance est suprieure  la rfrence.
   */
  private static int compareDistanceWithReference(double distance, double reference, double up_multiplicator, double down_multiplicator)
  {
    // Cas #1 : Si la distance mentionne  l'intrieure
    if(distance < down_multiplicator * reference)
      return -1;
    
    // Cas #2 : Si la distance mentionne  l'extrieur
    if(distance > up_multiplicator * reference)
      return 1;
     
    // Cas #3 : La distance est gale  la rfrence
    return 0;
  }
  
}
