/**
 * 
 */
package sim.geometry.space;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import sim.geometry.SGeometry;
import sim.geometry.SRay;
import sim.math.SVector3d;

/**
 * La classe abstraite <b>SAbstractVoxelSpace</b> reprsente un espace de gomtrie dont le partitionnement
 * de l'espace s'effectue par voxel. Cette classe contient des mthodes utilitaires pour analyser cet espace.
 * 
 * @author Simon Vzina
 * @since 2015-12-24
 * @version 2015-12-29
 */
public abstract class SAbstractVoxelSpace extends SAbstractGeometrySpace {

  //------------
  // VARIABLE //
  //------------
  
  /**
   * La variable 'linear_list' correspond  une liste de primitive ne fait pas partie de la carte des voxels puisque ce sont des primitives <b>sans bote englobante</b>.
   * L'algorithme d'intersection testera alors toutes les primitives de cette liste avec chaque rayon. 
   */
  protected List<SGeometry> linear_list;  
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un espace de voxel abstrait par dfaut. 
   */
  public SAbstractVoxelSpace()
  {
    super();
    
    linear_list = new ArrayList<SGeometry>();
  }

  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir l'intersection la plus prs entre un rayon et des gomtries situes dans une carte de voxels. 
   * 
   * @param ray - Le rayon  intersecter avec les gomtries.
   * @param t_max - Le temps maximal.
   * @param map - La carte des voxels.
   * @param builder - Le constructeur de voxel (pour vrifier si l'intersection est  l'intrieur du voxel).
   * @param voxel - Le voxel de la carte o sont raliss les tests d'intersection.
   * @return Le rayon avec les caractristiques de l'intersection (s'il y en a eu une) de temps minimum.
   */
  protected SRay nearestIntersectionInVoxelMap(SRay ray, double t_max, Map<SVoxel,List<SGeometry>> map, SVoxelBuilder builder, SVoxel voxel)
  {
    if(map != null)         // Si la carte n'est pas null
      if(!map.isEmpty())    // Si la carte n'est pas vide
      {
        // Obtenir la liste des gomtries en rfrence dans le voxel
        List<SGeometry> list = map.get(voxel);
        
        if(list != null)          // Si le voxel est dans la carte
          if(!list.isEmpty())     // Si la liste de gomtrie associe  ce voxel n'est pas vide
          {
            // Raliser les intersections avec la liste disponible dans le voxel.
            // Cette liste sera trie en ordre croissant de temps (le plus petit temps en premier dans la liste)
            List<SRay> list_intersection = intersections(list, ray, t_max);   
                
            // Regarder la liste des intersections et prendre l'intersection au temps le plus petit,
            // mais qui se retrouve dans le voxel courant (sinon, il est rejet).
            // Comparer celles trouves avec le rayon dj intersect pour garder celle de moindre temps
            if(!list_intersection.isEmpty())
              for(SRay r : list_intersection)
                if(builder.buildVoxel(r.getIntersectionPosition()).equals(voxel))   
                  return r;
          }
      }
    
    // Il n'y a pas d'intersection adquate qui a t ralis, on retourne un rayon sans intersection
    return ray;
  }
    
  /**
   * Mthode pour obtenir la liste des intersections transparente en ordre dcroissant dont la plus loign (premire de la liste) sera une gomtrie opaque s'il y a eu intersection de cette nature.
   * 
   * @param ray - Le rayon  intersecter.
   * @param t_max - Le temps maximal pouvant tre parcouru par le rayon.
   * @param map - La carte des voxels.
   * @param builder - Le constructeur de voxel (pour vrifier si l'intersection est  l'intrieur du voxel).
   * @param voxel - Le voxel de la carte o sont raliss les tests d'intersection.
   * @return La liste des intersections transparente en odre dcroissant dont le premier lment sera une gomtrie opaque s'il y a eu intersection de cette nature.
   */
  protected List<SRay> nearestOpaqueIntersectionInVoxelMap(SRay ray, double t_max, Map<SVoxel,List<SGeometry>> map, SVoxelBuilder builder, SVoxel voxel)
  {
    if(map != null)
      if(!map.isEmpty())
      {
        // Obtenir la liste des gomtries en rfrence dans le voxel
        List<SGeometry> list = map.get(voxel);
        
        if(list != null)          
          if(!list.isEmpty())     
            return nearestOpaqueIntersectionInVoxel(list, ray, t_max, builder, voxel);  // Obtenir la liste de l'intersection opaque la plus prs
      }

    // La carte n'tant vide, on retourne une liste vide
    return new ArrayList<SRay>();    
  }
  
  /**
   * <p>
   * Mthode qui value l'intersection la plus prs entre un rayon et les diffrentes gomtries <b>opaque</b> de l'espace ne dpassant pas un certain temps maximal.
   * Une liste des gomtries transparentes sera ordonne en <b>ordre inverse d'apparition</b> (ordre dcroissant du temps).
   * La <b>liste dbutera par une gomtrie opaque</b> s'il y a eu une intersection avec une gomtrie opaque.
   * S'il n'y a pas eu d'intersection avec une gomtrie opaque, elle contiendra uniquement des gomtries transparentes.
   * </p>
   * 
   * <p>
   * En comparaison avec sa version prcdente, cette mthode n'acceptera uniquement les intersections situes  l'intrieur du voxel associ  la liste des gomtries.
   * </p>
   *
   * @param list - La liste des gomtries.
   * @param ray - Le rayon  intersecter avec les gomtries de l'espace.
   * @param t_max - Le temps maximal.
   * @param builder - Le constructeur de voxel (pour vrifier si l'intersection est  l'intrieur du voxel).
   * @param voxel - Le voxel de la carte o sont raliss les tests d'intersection.
   * @return Une liste des gomtries transparentes intersectes en <b>ordre dcroissant</b> dbutant par une gomtrie opaque s'il y a eu lieu. La liste sera <b>vide</b> s'il n'y a <b>aucune intersection</b> sur le temps maximal.
   */
  private List<SRay> nearestOpaqueIntersectionInVoxel(List<SGeometry> list, SRay ray, final double t_max, SVoxelBuilder builder, SVoxel voxel)
  {
    // Obtenir la liste complte des intersections tries en ordre croissant
    List<SRay> list_intersection = intersections(list, ray, t_max);
    
    // Iterer sur l'ensemble des intersections et conserver celle 
    // o il y a des gomtries transparentes et une seule opaque s'il y en a une
    List<SRay> return_list = new ArrayList<SRay>();
    
    for(SRay r : list_intersection)
    {
      // Ajouter et traiter l'intersection uniquement si elle se retrouve dans le voxel
      if(builder.buildVoxel(r.getIntersectionPosition()).equals(voxel))
      {
        return_list.add(r);   // ajouter cette intersection
        
        // Si la gomtrie intersect n'est pas transparente,
        // ce sera la dernire intersection ajoute  la liste
        // car elle sera la seule intersection opaque de la liste.
        if(!r.getGeometry().isTransparent())
        {
          Collections.reverse(return_list);     // Inverser l'ordre en ordre dcroissant
          return return_list;                   // retourner une liste avec gomtrie opaque
        }
      }
    }
    
    Collections.reverse(return_list);           // Inverser l'ordre en ordre dcroissant
    return return_list;                         // retourner une liste sans gomtrie opaque
  }
  
  /**
   * Mthode qui value l'ensemble des gomtries d'une carte de voxel o une vecteur v se retrouve  l'intrieur.
   * 
   * @param map - La carte des voxels.
   * @param builder - Le constructeur de voxel attitr  la carte de voxel.
   * @param v - La position du vecteur.
   * @return La liste des gomtries o le vecteur v se retrouve  l'intrieur.
   */
  protected List<SGeometry> listInsideGeometryInMap(Map<SVoxel,List<SGeometry>> map, SVoxelBuilder builder, SVector3d v)
  {
    // Obtenir la liste des gomtries dans la cartes des voxel 
    // et obtenir les gomtries o le vecteur v s'y retrouve
    if(map != null)        
      if(!map.isEmpty())    
        return listInsideGeometry(map.get(builder.buildVoxel(v)), v);
        
    // La carte tant null ou vide, nous retournons une liste vide.
    return new ArrayList<SGeometry>();
  }
  
  /**
   * Mthode pour ajouter des gomtries  la carte de voxel avec les voxels o la gomtrie est situe.
   * 
   * @param map - La carte des voxels contenant une liste de gomtrie.
   * @param extremum_voxel - Le voxel extremum associ  la carte de voxels.
   * @param geometry - La gomtrie  ajouter.
   * @param list - La liste des voxels o la gomtrie est situe.
   */
  protected SVoxel addGeometryToMap(Map<SVoxel, List<SGeometry>> map, SVoxel extremum_voxel, SGeometry geometry, List<SVoxel> list)
  {
    // Iterer sur la liste des voxels
    for(SVoxel v : list)
    {  
      // Si le voxel est dj dans la carte : ajouter la gomtrie  la liste existante de gomtrie de ce voxel
      if(map.containsKey(v))            
        map.get(v).add(geometry);
      else                                    
      {
        // Si le voxel n'est pas dans la carte : crer une liste avec la gomtrie dedans. 
        // Ajouter le voxel comme cl de recherche dans la carte.
        // Il faudra ajuster le voxel d'extrme galement.
        List<SGeometry> l = new ArrayList<SGeometry>();
        
        // Ajouter la gomtrie  la liste
        l.add(geometry);                                       
        
        // Ajouter le voxel  la carte
        map.put(v, l);                                           
        
        // Mise  jour du voxel d'extremum (si ncessaire)
        extremum_voxel = updateExtremumVoxel(extremum_voxel, v);   
      }
    }
    
    return extremum_voxel;
  }
  
  /**
   * Mthode pour faire la mise  jour du voxel dfinissant les extermums en valeur absolue d'une carte de voxels.
   * 
   * @param previous_extremum - Le voxel tant celui dfinissant l'ancien extremum.
   * @param new_voxel - Le nouveau voxel pouvant redfinir le voxel extremum.
   * @return Un nouveau voxel dfinissant un nouvel extremum ou celui pass en paramtre s'il demeure extremum.
   */
  protected SVoxel updateExtremumVoxel(SVoxel previous_extremum, SVoxel new_voxel)
  {
    // Voxel de retour tant l'ancien par dfaut (s'il n'est pas chang)
    SVoxel return_voxel = previous_extremum;
    
    //Mise  jour si ncessaire de la coordonne extermum en x
    if(Math.abs(new_voxel.getX()) > return_voxel.getX())
      return_voxel = new SVoxel(Math.abs(new_voxel.getX()), return_voxel.getY(), return_voxel.getZ());
    
    //Mise  jour si ncessaire de la coordonne extermum en y
    if(Math.abs(new_voxel.getY()) > return_voxel.getY())
      return_voxel = new SVoxel(return_voxel.getX(), Math.abs(new_voxel.getY()), return_voxel.getZ());
    
    //Mise  jour si ncessaire de la coordonne extermum en z
    if(Math.abs(new_voxel.getZ()) > return_voxel.getZ())
      return_voxel = new SVoxel(return_voxel.getX(), return_voxel.getY(), Math.abs(new_voxel.getZ()));
    
    return return_voxel;
  }
  
  /**
   * Mthode pour obtenir le nombre de gomtries en rfrence dans les voxels.
   * 
   * @param map - La carte des voxels.
   * @return Le nombre de primitives dans les voxels de la carte.
   */
  protected int evaluateNbGeometryReference(Map<SVoxel, List<SGeometry>> map)
  {
    int count = 0;
    
    // Iterer sur l'ensemble des valeurs de la carte
    for(List<SGeometry> l : map.values())
      count += l.size();
    
    return count;
  }
  
  /**
   * Mthode pour obtenir le nombre de gomtries en rfrence en moyenne par voxel.
   * 
   * @param map - La carte des voxels.
   * @return Le nombre moyen de primitives par voxel.
   */
  protected double evaluateNbGeometryReferencePerVoxel(Map<SVoxel, List<SGeometry>> map)
  {
    // Valeur statistique tant un calcul de moyenne
    return (double) evaluateNbGeometryReference(map) / (double) map.keySet().size();
  }
  
}//fin de la classe SAbstractVoxelSpace
