/**
 * 
 */
package sim.geometry.space;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import sim.exception.SConstructorException;
import sim.geometry.SGeometry;
import sim.math.SStatistic;

/**
 * La classe <b>SGeometryCollectionSplitter</b> permet de sparer en sous-groupe une collection de gomtrie.
 * Ces gomtries sont regroup en fonction de la taille des <b>botes englobantes</b> les recouvrant.
 * 
 * @author Simon Vzina
 * @since 2015-12-15
 * @version 2016-02-14
 */
public class SGeometryCollectionSplitter {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>DEFAULT_RECURSIVITY_LEVEL</b> reprsente le niveau de rcursivit de certains algorithmes par dfaut. 
   */
  private static final int DEFAULT_RECURSIVITY_LEVEL = 1;
  
  /**
   * La constante <b>MINIMUM_GEOMETRY_PER_LIST</b> reprsente le nombre minimal de gomtrie pouvant tre situe dans une liste.
   * Ainsi, une liste ayant moins de gomtrie que la valeur {@value} ne devrait pas tre scinde.
   */
  private static final int MINIMUM_GEOMETRY_PER_LIST = 100;
  
  //Les codes de rfrence des diffrents algorithmes disponibles
  public static final int SPLIT_BOX_AND_NO_BOX  = 0;
  public static final int SPLIT_HALF_AND_HALF = 1;
  public static final int SPLIT_AT_AVERAGE_SIZE = 2;
  public static final int SPLIT_AT_HALF_SIGMA = 3;
    
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>box_builder</b> correspond au constructeur de bote englobante des gomtries.
   */
  private final SBoundingBoxBuilder box_builder;
  
  /**
   * La variable <b>bounding_box_split_list</b> correspond  la liste de listes de botes englobante ayant t spares.
   * Ces botes englobantes contiennent les gomtries qui doivent tre organises spatialement. 
   */
  private final List<List<SBoundingBox>> bounding_box_split_list;
  
  /**
   * La variable <b>geometry_no_box_list</b> correspond  la liste de gomtries n'ayant pas de bote englobante.
   * Ces gomtries ne se retrouveront donc pas dans la liste <b>geometry_split_list</b>.
   */
  private final List<SGeometry> geometry_no_box_list;
  
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * Constructeur d'un sparateur de collection de gomtries  niveau de rcursivit par dfaut.
   * 
   * @param list - La liste de gomtrie  sparer.
   * @param split_code - Le code de l'algorithme de sparation utilis par le sparateur.
   * @throws SConstructorException Si le code de sparation (split_code) n'est pas valide.
   */
  public SGeometryCollectionSplitter(List<SGeometry> list, int split_code) throws SConstructorException
  {
    this(list, split_code, DEFAULT_RECURSIVITY_LEVEL);
  }

  /**
   * Constructeur d'un sparateur de collection de gomtries.
   * 
    *@param list - La liste de gomtrie  sparer.
   * @param split_code - Le code de l'algorithme de sparation utilis par le sparateur.
   * @param recursivity_level
   * @throws SConstructorException Si le code de sparation (split_code) n'est pas valide.
   */
  public SGeometryCollectionSplitter(List<SGeometry> list, int split_code, int recursivity_level) throws SConstructorException
  {
    box_builder = new SBoundingBoxBuilder();
    bounding_box_split_list = new ArrayList<List<SBoundingBox>>();
    
    geometry_no_box_list = new ArrayList<SGeometry>();
    
    // Appliquer l'algorithme de sparation demand
    switch(split_code)
    {
      case SPLIT_BOX_AND_NO_BOX :       splitBoxAndNoBox(list); break;
      
      case SPLIT_HALF_AND_HALF :        splitHalfAndHalf(list); break;
      
      case SPLIT_AT_AVERAGE_SIZE :      splitAtAverageSize(list, recursivity_level); break;
      
      case SPLIT_AT_HALF_SIGMA :     splitAtHalfSigma(list, recursivity_level); break;
      
      default : throw new SConstructorException("Erreur SGeometryCollectionSplitter 001 : Le code de rfrence d'algorithme " + split_code + " n'est pas reconnu.");
    }
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir la liste des listes de gomtries aprs le fractionnement de l'ensemble.
   * 
   * @return La liste des listes de gomtries.
   */
  public List<List<SGeometry>> getSplitList()
  {
    List<List<SGeometry>> list_of_list = new ArrayList<List<SGeometry>>();
    
    // Convertir une liste de liste de botes englobante en liste de liste de gomtries 
    for(List<SBoundingBox> l : bounding_box_split_list)
      list_of_list.add(fillGeometryList(l));
    
    return list_of_list;
  }
  
  /**
   * Mthode pour otenir la liste des listes de botes englobantes aprs le frationnement de l'ensemble.
   * 
   * @return La liste des listes de botes englobantes.
   */
  public List<List<SBoundingBox>> getBoundingBoxSplitList()
  {
    return bounding_box_split_list;
  }
  
  /**
   * Mthode pour obtenir la liste des gomtries n'tant pas fractionnes dans les ensembles
   * puisque ces gomtries ne possde pas de bote englobante.
   * 
   * @return La liste des gomtries sans bote englobante.
   */
  public List<SGeometry> getNoBoxList()
  {
    return geometry_no_box_list;
  }
  
  /**
   * Mthode qui effectue le filtrage des gomtries ne pouvant obtenir de bote englobante avec celle pouvant obtenir les botes englobantes.
   * Cette mthode va intgrer les gomtries <b>sans botes</b>  la liste <b>geometry_no_box_list</b>. 
   * Les autres gomtries seront retournes sous le format d'une une liste de bote englobante o un triage particulier subsquent sera effectu au besoin.  
   * 
   * @param list La liste des gomtries  filtrer.
   * @return Une liste de botes englobantes.
   */
  private List<SBoundingBox> filterBoxAndNoBox(List<SGeometry> list)
  {
    List<SBoundingBox> list_box = new ArrayList<SBoundingBox>();
    
    // Remplir la liste de botes englobantes et la liste des gomtries sans botes englobantes
    for(SGeometry g : list)
    {
      SBoundingBox box = box_builder.buildBoundingBox(g);
      
      if(box != null)
        list_box.add(box);
      else
        geometry_no_box_list.add(g);
    }
    
    return list_box;
  }
  
  /**
   * Mthode qui effectue la sparation d'une liste de gomtries en deux catgories : 
   * <ul>- avec bote englobante</ul>
   * <ul>- sans bote englobante</ul>
   * 
   * @param list - La liste des gomtries  sparer.
   */
  private void splitBoxAndNoBox(List<SGeometry> list)
  {
    List<SBoundingBox> list_box = filterBoxAndNoBox(list);    // obtenir la liste des botes
    
    bounding_box_split_list.add(list_box);
  }
  
  /**
   * Mthode qui effectue la sparation d'une liste de gomtries en deux ensembles gaux tel que
   * les plus petites botes englobantes seront dans la 1ire liste et les plus grandes botes
   * englobantes seront dans la 2ime liste.
   * 
   * @param list - La liste des gomtries  sparer.
   */
  private void splitHalfAndHalf(List<SGeometry> list)
  {
    List<SBoundingBox> list_box = filterBoxAndNoBox(list);    // obtenir la liste des botes
    
    // Trier la liste des botes englobantes en fonction de leur taille (selon l'implmentation de comparable) 
    Collections.sort(list_box);
    
    List<SBoundingBox> list_1 = new ArrayList<SBoundingBox>();
    List<SBoundingBox> list_2 = new ArrayList<SBoundingBox>();
    
    int middle = list_box.size() / 2;
    
    // Remplir la 1ire liste
    for(int i = 0; i < middle; i++)
      list_1.add(list_box.get(i));
    
    // Remplir la 2ime liste
    for(int i = middle; i < list_box.size(); i++)
      list_2.add(list_box.get(i));
        
    bounding_box_split_list.add(list_1);
    bounding_box_split_list.add(list_2);
  }
    
  /**
   * Mthode pour sparer une liste de gomtries en deux sous-liste  la valeur moyenne de la taille maximale des botes englobante.
   * Si le niveau de rcursivit est suprieur  2, les deux sous-liste seront  nouveau spares par le mme procd.
   * 
   * @param list - La liste des gomtries  sparer.
   * @param level - Le niveau de rcursivit que la liste doit tre spar en deux.
   */
  private void splitAtAverageSize(List<SGeometry> list, int level)
  {
    // Obtenir la liste de bote englobante
    List<SBoundingBox> list_box = filterBoxAndNoBox(list);
      
    // Sparer la liste des gomtries situes dans la liste des botes
    if(!list_box.isEmpty())
      splitBoxListAtAverageSize(list_box, level);      
  }
  
  /**
   * Mthode pour sparer une liste de gomtries en deux sous-liste  la valeur moyenne de la taille maximale des botes englobante.
   * Si le niveau de rcursivit est suprieur  2, les deux sous-liste seront  nouveau spares par le mme procd.
   * 
   * @param list_box - La liste des botes contenant des gomtries  sparer.
   * @param level - Le niveau de rcursivit que la liste doit tre spar en deux.
   */
  private void splitBoxListAtAverageSize(List<SBoundingBox> list_box, int level)
  {
    // Sparer la liste si :
    // - le niveau de rcursivit est suprieur  1 
    // - Si le nombre de botes dans la liste est suprieur  un seuil minimal
    if(level > 1 && list_box.size() >  MINIMUM_GEOMETRY_PER_LIST)
    {
      // valuer la taille moyenne des botes
      double average = SStatistic.average(list_box);
      
      // Sparer en deux listes : Infrieur  la moyenne et suprieur  la moyenne
      List<SBoundingBox> list_down = new ArrayList<SBoundingBox>();
      List<SBoundingBox> list_up = new ArrayList<SBoundingBox>();
      
      for(SBoundingBox b : list_box)
        if(b.getStatisticalValue() < average)
          list_down.add(b);
        else
          list_up.add(b);
      
      // Faire  nouveau une sparation avec les deux nouvelles listes
      splitBoxListAtAverageSize(list_down, level-1);
      splitBoxListAtAverageSize(list_up, level-1);
    }
    else
    {
      // Mettre les botes contenues dans la liste dans la liste des listes de botes
      bounding_box_split_list.add(list_box);
    }
  }
    
  /**
   * Mthode pour sparer une liste de gomtries en trois sous-liste en utilisant la <b>moyenne</b> et <b>l'cart-type</b>.
   * Si le niveau de rcursivit est suprieur  2, les sous-liste seront  nouveau spares par le mme procd.
   * 
   * @param list - La liste des gomtries  sparer.
   * @param level - Le niveau de rcursivit que la liste doit tre spar en deux.
   */
  private void splitAtHalfSigma(List<SGeometry> list, int level)
  {
    // Obtenir la liste de bote englobante
    List<SBoundingBox> list_box = filterBoxAndNoBox(list);
      
    // Sparer la liste des gomtries situes dans la liste des botes
    if(!list_box.isEmpty())
      splitBoxListAtHalfSigma(list_box, level);      
  }
  
  /**
   * Mthode pour sparer une liste de gomtries en trois sous-liste en utilisant la <b>moyenne</b> et <b>l'cart-type</b>.
   * Si le niveau de rcursivit est suprieur  2, les sous-liste seront  nouveau spares par le mme procd.
   * 
   * @param list_box - La liste des botes contenant les gomtries  sparer.
   * @param level - Le niveau de rcursivit que la liste doit tre spar en deux.
   */
  private void splitBoxListAtHalfSigma(List<SBoundingBox> list_box, int level)
  {
    // Sparer la liste si :
    // - le niveau de rcursivit est suprieur  1 
    // - Si le nombre de botes dans la liste est suprieur  un seuil minimal
    if(level > 1 && list_box.size() >  MINIMUM_GEOMETRY_PER_LIST)
    {
      // valuer les paramtres statistiques
      double average = SStatistic.average(list_box);
      double sigma = SStatistic.standardDeviation(list_box);
      
      double min_value = average - sigma/2.0;
      double max_value = average + sigma/2.0;
      
      // Sparer en trois listes : 
      List<SBoundingBox> list_down = new ArrayList<SBoundingBox>();
      List<SBoundingBox> list_center = new ArrayList<SBoundingBox>();
      List<SBoundingBox> list_up = new ArrayList<SBoundingBox>();
      
      for(SBoundingBox b : list_box)
        if(b.getStatisticalValue() < min_value)
          list_down.add(b);
        else
          if(b.getStatisticalValue() > max_value)
            list_up.add(b);
          else
            list_center.add(b);
      
      // Faire  nouveau une sparation avec les deux listes down et up
      splitBoxListAtHalfSigma(list_down, level-1);
      splitBoxListAtHalfSigma(list_up, level-1);
      
      // Mettre les botes contenues dans la liste du centre dans la liste des listes de botes
      bounding_box_split_list.add(list_center);
    }
    else
    {
      // Mettre les botes contenues dans la liste dans la liste des listes de botes
      bounding_box_split_list.add(list_box);
    }
  }
  
  /**
   * Mthode qui fait la construction d'une liste de gomtries  partir d'une liste de botes englobantes
   * o une rfrence  une gomtrie s'y retrouve.
   * 
   * @param list - La liste des botes englobantes.
   * @return La liste des gomtries.
   */
  private List<SGeometry> fillGeometryList(List<SBoundingBox> list)
  {
    List<SGeometry> geometry_list = new ArrayList<SGeometry>();
    
    for(SBoundingBox b : list)
      geometry_list.add(b.getGeometry());
    
    return geometry_list;
  }
  
}//fin de la classe SGeometryCollectionSplitter
