/**
 * 
 */
package sim.graphics.light;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.math.SVector3d;
import sim.physics.SNotVisibleLightException;
import sim.physics.SWave;
import sim.physics.SWaveOptics;
import sim.readwrite.SKeyWordDecoder;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe abstraite <b>SAbstractInterferenceLight</b> correspond  une source de lumire monochromatique pouvant raliser de l'interfrence.
 * 
 * @author Simon Vzina
 * @since 2016-02-15
 * @version 2017-02-02
 */
public abstract class SAbstractInterferenceLight extends SAbstractAttenuatedLight implements SInterferenceLight {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = {
    SKeyWordDecoder.KW_WAVE_LENGTH, SKeyWordDecoder.KW_PERIOD_ITERATION, SKeyWordDecoder.KW_NB_OSCILLATOR
  };
  
  /**
   * La constante <b>DEFAULT_WAVE_LENGTH</b> correspond  la longueur d'onde (en nm) de la source de lumire monochromatique par dfaut.
   */
  protected static final double DEFAULT_WAVE_LENGTH = 550;
  
  /**
   * La constante <b>DEFAULT_PERIOD_ITERATION</b> correspond au nombre d'itration qui seront ralis lors d'un calcul effectu sur une priode complte d'oscillation des oscillateurs.
   */
  protected static final int DEFAULT_PERIOD_ITERATION = 10;
  
  /**
   * La constante <b>DEFAULT_NB_OSCILLATOR</b> correspond au nombre d'oscillateurs dfinissant un axe de la fente par dfaut.
   */
  protected static final int DEFAULT_NB_OSCILLATOR = 2;
  
  //-------------
  // VARIABLES //
  //-------------
  /**
   * La variable <b>wave_length</b> correspond  la longueur d'onde de la source de lumire.
   */
  protected double wave_length;
  
  /**
   * La variable <b>period_iteration</b> correspond au nombre d'itrations qui seront utilises pour valuer une priode complte 
   * d'oscillation des oscillateurs composant la source de lumire. Ce paramtre dtermine la prcision lorsqu'il
   * y aura lors d'un calcul qui require une moyenne sur cylce complet d'oscillation. 
   */
  protected int period_iteration;
  
  /**
   * La variable <b>nb_oscillator</b> correspond au nombre d'oscillateurs utilisa pour dfinir l'axe vertical (le <i>up</i>) de l'ouverture.
   */
  protected int nb_oscillator;
  
  /**
   * La variable <b>wave_list</b> correspond  la liste ondes progressives gnr par les multiples oscillateurs internes  la source de lumire.
   * Ce sont ces oscillateurs qui vont gnrer des ondes qui pourront interfrer au site o la lumire se dirigera.
   */
  protected final List<SWave> wave_list;
  
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * Constructeur d'une source de lumire pouvant raliser de l'interfrence par dfaut. 
   */
  public SAbstractInterferenceLight()
  {
    this(DEFAULT_WAVE_LENGTH, DEFAULT_POSITION, DEFAULT_AMPLIFICATION, DEFAULT_CONSTANT_ATTENUATOR, DEFAULT_LINEAR_ATTENUATOR, DEFAULT_QUADRATIC_ATTENUATOR, DEFAULT_PERIOD_ITERATION, DEFAULT_NB_OSCILLATOR);
  }

  /**
   * Constructeur d'une source de lumire pouvant raliser de l'interfrence.
   * 
   * @param wave_length La longueur d'onde de la source de lumire (en nm).
   * @param position La position de la source.
   * @param amp Le facteur d'amplification de la source de lumire.
   * @param cst_att La constante d'attnuation  taux constant.
   * @param lin_att La constante d'attnuation  taux linaire.
   * @param quad_att La constante d'attnuation  taux quadratique.
   * @param period_iteration Le nombre d'itrations ralises dans les calculs effectus sur un cycle de priode d'oscillation des oscillateurs.
   * @throws SConstructorException Si une erreur de construction est survenue.
   */
  public SAbstractInterferenceLight(double wave_length, SVector3d position, double amp, double cst_att, double lin_att, double quad_att, int period_iteration, int nb_oscillator) throws SConstructorException
  {
    super(DEFAULT_LIGHT_COLOR, position, amp, cst_att, lin_att, quad_att);
    
    if(wave_length < 0)
      throw new SConstructorException("Erreur SAbstractInterferenceLight 001 : La longueur d'onde de la source de lumire doit tre positive.");
    
    if(period_iteration < 1)
      throw new SConstructorException("Erreur SAbstractInterferenceLight 002 : Le nombre d'itration sur une priode tant de '" + period_iteration + "' doit tre suprieur  1.");
    
    if(nb_oscillator < 1)
      throw new SConstructorException("Erreur SAbstractInterferenceLight 003 : Le nombre d'oscillateur pour un axe de l'ouverture '" + nb_oscillator + "' doit tre suprieur  1.");
    
    this.wave_length = wave_length;
    
    this.period_iteration = period_iteration;
    
    this.nb_oscillator = nb_oscillator;
    
    wave_list = new ArrayList<SWave>();
    
    try{
      initialize();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SAbstractInterferenceLight 004 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }
  }
 
  
  
  /**
   * Mthode pour crire les paramtres associs  la classe SAbstractInterferenceLight et ceux qu'il hrite.
   * 
   * @param bw Le BufferedWriter crivant l'information dans un fichier txt.
   * @throws IOException S'il y a une erreur d'criture.
   * @see IOException
   */
  protected void writeSAbstractInterferenceLightParameter(BufferedWriter bw)throws IOException
  {
    //crire les paramtres hrits de la classe SAbstractAttenuatedLight
    writeSAbstractAttenuatedLightParameter(bw);
    
    bw.write(SKeyWordDecoder.KW_WAVE_LENGTH);
    bw.write("\t\t");
    bw.write(Double.toString(wave_length));
    bw.write(" nm");
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_PERIOD_ITERATION);
    bw.write("\t\t");
    bw.write(Integer.toString(period_iteration));
    bw.write(SStringUtil.END_LINE_CARACTER);
    
    bw.write(SKeyWordDecoder.KW_NB_OSCILLATOR);
    bw.write("\t\t");
    bw.write(Integer.toString(nb_oscillator));
    bw.write(SStringUtil.END_LINE_CARACTER);
  }
  
  @Override
  public double getRelativeIntensity(SVector3d position_to_illuminate)
  {
    // Vrifier que la liste des ondes n'est pas vide
    if(wave_list.isEmpty())
      return 0.0;
    
    // valuer la priode sur laquelle la moyenne sera effectue (le temps d'une oscillation complte)
    double period = wave_list.get(0).getOscillator().getPeriod();
    
    // valuer l'intensit moyenne sur un cycle complet des oscillateur
    double intensity = SWaveOptics.interferenceAverageWaveValue(wave_list, position_to_illuminate, period, period_iteration);
    
    if(intensity < 0)
      throw new SRuntimeException("SAbstractInterferenceLight --> intensity = " + intensity);
    
    // Nombre d'oscillateurs utiliss dans le calcul
    //int nb = wave_list.size();                        // ce choix va crer par moment des valeurs ngative, ce qui n'est pas mathmatiquement possible.
    double nb = (double)wave_list.size();               // ce choix semble rgler le problme !
    
    //--------------------------------
    // CODE DE TEST --->  RETIRER !!!
    //--------------------------------
    if(intensity/(nb*nb) < 0)
      throw new SRuntimeException("SAbstractInterferenceLight::public double getRelativeIntensity(SVector3d position_to_illuminate) --> intensity/(nb*nb) = " + intensity/(nb*nb) + ", intensity = " + intensity + ", nb = " + nb);
    
    // On retourne l'intensit relative au maximum de l'interfrence constructive 
    // tant gale au nombre d'oscillateur au carr 
    return intensity / (nb*nb);
  }

  /**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
  {
    // Vider la liste existante (pouvant tre construite lors d'une constructure, mais qui devra tre change aprs une lecture)
    wave_list.clear();
    
    // Dterminer la couleur associe  la longueur d'onde
    try{   
      color = SWaveOptics.wavelengthToSColor(wave_length);
    }catch(SNotVisibleLightException e){
      throw new SInitializationException("Erreur SAbstractInterferenceLight 005 : La source de lumire n'est pas dans le visible. On ne peut pas dfinir de couleur." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }
  }
  
  @Override
  protected void readingInitialization() throws SInitializationException
  {
    super.readingInitialization();
    
    initialize();
  }
  
  @Override
  public String[] getReadableParameterName()
  {
    String[] other_parameters = super.getReadableParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException 
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_WAVE_LENGTH : wave_length = readDoubleGreaterThanZero(remaining_line, SKeyWordDecoder.KW_WAVE_LENGTH); return true;
     
      case SKeyWordDecoder.CODE_PERIOD_ITERATION : period_iteration = this.readIntGreaterThanZero(remaining_line, SKeyWordDecoder.KW_PERIOD_ITERATION); return true;
      
      case SKeyWordDecoder.CODE_NB_OSCILLATOR : nb_oscillator = readIntEqualOrGreaterThanZero(remaining_line, SKeyWordDecoder.KW_NB_OSCILLATOR); return true;
      
      // Lecture des paramtres de la classe hrit SAbstractAttenuatedLight  
      default : return super.read(sbr, code, remaining_line);
    }
  }
    
}//fin de la classe SAbstractInterferenceLight
