/**
 * 
 */
package sim.graphics.material;

import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.graphics.SColor;
import sim.util.SInitializationException;
import sim.util.SStringUtil;

/**
 * Classe qui reprsente un matriel respectant les standards de la librairie OpenGl. Par dfaut, la couleur du
 * matriel est blanc (1.0, 1.0, 1.0) ou la couleur de la texture et on lui appliquer un vecteur de coefficient
 * de rflexion (ka, kd, ks) qui aura pour effet de filtrer la couleur de base du matriel selon le modle d'illumination.
 * 
 * @author Simon Vzina
 * @since 2015-01-13
 * @version 2015-12-05
 */
public class SOpenGLMaterial extends SDefaultMaterial {

  //Couleur de base d'un matriel OpenGl (couleur qui devra tre multiplie par les coefficients de rflexion)
  protected static final SColor COLOR_BASE = new SColor(1.0, 1.0, 1.0);           
  
  //Paramtre par dfaut
  private static final SColor DEFAULT_AMBIENT_COEFFICIENT = new SColor(0.1, 0.1, 0.1);  
  private static final SColor DEFAULT_DIFFUSE_COEFFICIENT = new SColor(0.3, 0.7, 0.3);      //sera vert !!!
  private static final SColor DEFAULT_SPECULAR_COEFFICIENT = new SColor(0.05, 0.05, 0.05);
  private static final double DEFAULT_SHININESS = 5;
  
  private static final String DEFAULT_NAME = "default_openGL_material"; //nom par dfaut
  
  private final SColor coefficient_ambient;   //vecteur des coefficients de rflexion ambiant   
  private final SColor coefficient_diffuse;   //vecteur des coefficients de rflexion diffus
  private final SColor coefficient_specular;  //vecteur des coefficients de rflexion spculaire
  private final double shininess;             //Niveau de brillance du matriel
  
  /**
   * Constructeur d'un matriel respectant les standards de la librairie OpenGL blanc avec nom par dfaut.
   */
  public SOpenGLMaterial()
  {
    this(DEFAULT_NAME);
  }
  
  /**
   * Constructeur d'un matriel respectant les standards de la librairie OpenGL blanc.
   * @param name - Le nom du matriel.
   */
  public SOpenGLMaterial(String name)
  {
    this(name, DEFAULT_AMBIENT_COEFFICIENT, DEFAULT_DIFFUSE_COEFFICIENT, DEFAULT_SPECULAR_COEFFICIENT, DEFAULT_SHININESS);
  }
  
  /**
   * Constructeur d'un matriel respectant les standards de la librairie OpenGL.
   * @param name - Le nom du matriel.
   * @param ka - Le vecteur des coefficients de rflexion ambiant.
   * @param kd - Le vecteur des coefficients de rflexion diffuse.
   * @param ks - Le vecteur des coefficients de rflexion spculaire.
   * @param shininess - La brillance du matriel.
   * @throws SConstructorException Si la brillance du matriel est ngative.
   * @throws SConstructorException Si une erreur est survenue lors de la construction. 
   */
  public SOpenGLMaterial(String name, SColor ka, SColor kd, SColor ks, double shininess) throws SConstructorException
  {
    super(name);
    
    if(shininess < 0)
      throw new SConstructorException("Erreur SOpenGLMaterial 001 : La brillance (shininess) '" + shininess + "' est infrieur  zro.");
    
    coefficient_ambient = ka;
    coefficient_diffuse = kd;
    coefficient_specular = ks;
    this.shininess = shininess;
    
    try{
      initialize();
    }catch(SInitializationException e){
      throw new SConstructorException("Erreur SOpenGLMaterial 002 : Une erreur d'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
    }
  }
  
  /* (non-Javadoc)
   * @see simGraphic.SMaterial#ambientColor()
   */
  @Override
  public SColor ambientColor()
  { 
    return COLOR_BASE.multiply(coefficient_ambient);
  }
  
  /* (non-Javadoc)
   * @see simGraphic.SMaterial#diffuseColor()
   */
  @Override
  public SColor diffuseColor()
  { 
    return COLOR_BASE.multiply(coefficient_diffuse);
  }

  /* (non-Javadoc)
   * @see simGraphic.SMaterial#specularColor()
   */
  @Override
  public SColor specularColor()
  { 
    return COLOR_BASE.multiply(coefficient_specular);
  }
  
  /* (non-Javadoc)
	 * @see simGraphic.SMaterial#shininess()
	 */
	@Override
	public double getShininess()
	{ 
	  return shininess;
	}
	
	/**
   * Mthode pour faire l'initialisation de l'objet aprs sa construction.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  private void initialize() throws SInitializationException
  {
    
  }
  
	@Override
  protected void readingInitialization() throws SInitializationException
  {
    super.readingInitialization();
    
    initialize();
  }
	
	@Override
  public String getReadableName()
  {
    throw new SNoImplementationException("Erreur SOpenGLMaterial 003 : La mthode n'a pas t implmente.");
  }
	
}//fin classe SOpenGLMaterial
