/**
 * 
 */
package sim.math;

import sim.exception.SIllegalNegativeValueException;
import sim.exception.SNoImplementationException;
import sim.exception.SRuntimeException;

/**
 * La classe <b>SMath</b> contient des mthodes de calcul qui sont complmentaire  la classe java.lang.Math. 
 * Elle perment entre autre d'valuer les racines relles d'un polynme de degr 1, 2, 3 et 4.
 * 
 * @author Simon Vzina
 * @since 2015-02-19
 * @version 2020-01-08 (Version : Le ray tracer v2.107)
 */
public final class SMath {

  /**
   * La constante <b>NO_ROOT_SOLUTION</b> reprsente le tableau des solutions au racine d'un polynme o il n'y a <u>pas de solution relle</u>.
   * Le tableau est donc vide et de taille gale  zro.
   */
  public final static double[] NO_ROOT_SOLUTION = new double[0];
  
	/**
	 * <p>
	 * La constante <b>EPSILON</b> reprsentante un nombre trs petit positif, mais non nul. Ce chiffre peut tre utilis 
	 * pour comparer une valeur de type double avec le chiffre zro. Puisqu'un double gale  zro
	 * est difficile  obtenir numriquement aprs un calcul (sauf si l'on multiplie par zro), il est prfrable de 
	 * comparer un "pseudo zro" avec cette constante.
	 * </p>
	 * <p>
	 * Avec une valeur de EPSILON = 1e-10, cette valeur permet de comparer adquatement des nombres autour de '1' avec suffisamment de chiffres significatifs.
	 * </p>
	 */
	public static double EPSILON = 1e-10;           
	
	/**
	 * La constante <b>EPSILON_RELAXED</b> reprsente un nombre trs petit, mais mille fois plus grand que EPSILON (1000*EPSILON).
	 * Pour des raisons numriques, le seuil EPSILON est trop petit et ce nouveau seuil peut tre adquat. 
	 */
	public final static double EPSILON_RELAXED = EPSILON*1000.0;
	
	/**
   * La constante <b>NEGATIVE_EPSILON</b> reprsentante un nombre trs petit, mais non nul qui est <b>negatif</b>. Ce chiffre peut tre utilis 
   * pour comparer une valeur arbiraire de type double avec le chiffre zro, mais qui sera ngatif. Puisqu'un double gale  zro
   * est difficile  obtenir numriquement aprs un calcul (sauf si l'on multiplie par zro), il est prfrable de 
   * comparer un "pseudo zro" avec cette constante.
   */
	public final static double NEGATIVE_EPSILON = -1.0*EPSILON;  
	
	/**
	 * La constante <b>ONE_PLUS_EPSILON</b> reprsente une constante gale  <b>1 + EPSILON</b> ce qui correspond  un nombre lgrement suprieur  1.
	 */
	public final static double ONE_PLUS_EPSILON = 1 + EPSILON;
	
	/**
	 * La constante <b>ONE_MINUS_EPSILON</b> reprsente une constant gale  <b>1 - EPSILON</b> ce qui correspond  un nombre lgrement infrieur  1.
	 */
	public final static double ONE_MINUS_EPSILON = 1 - EPSILON;
	
	/**
   * La constante <b>ONE_PLUS_1000EPSILON</b> reprsente une constante gale  <b>1 + 1000*EPSILON</b> ce qui correspond  un nombre lgrement suprieur  1.
   */
  public final static double ONE_PLUS_1000EPSILON = 1 + 1000*EPSILON;
  
  /**
   * La constante <b>ONE_MINUS_EPSILON</b> reprsente une constant gale  <b>1 - 1000*EPSILON</b> ce qui correspond  un nombre lgrement infrieur  1.
   */
  public final static double ONE_MINUS_1000EPSILON = 1 - 1000*EPSILON;
  
	/**
	 * La constante <b>INFINITY</b> reprsente un nombre positif gale  l'infini. Cette valeur est obtenue 
	 * partir de la classe java.lang.Double.
	 * @see java.lang.Double
	 */
	public final static double INFINITY = Double.POSITIVE_INFINITY;
	
	/**
   * Mthode pour dterminer si deux nombres de type double sont <b>relativement gaux</b>. 
   * En utilisant une approche de calcul de diffrence, on vrifie si
   * <ul>a - b < EPSILON*ref</ul>  
   * afin de <b>valid l'galit</b> entre a et b (a == b). EPSILON est un seuil de prcision 
   * et ref est une base de rfrence (la valeur absolue la plus leve parmis a et b). 
   * <p>Cependant, si les deux chiffres sont infrieurs  EPSILON, ils seront considrs comme gaux.</p>
   * 
   * @param a Le 1ier nombre  comparer.
   * @param b Le 2ime nombre  comparer.
   * @return <b>true</b> si les deux nombres sont <b>relativement gaux</b> et <b>false</b> sinon.
   */
	public static boolean nearlyEquals(double a, double b)
	{
	  return nearlyEquals(a, b, EPSILON);
	}
	
	/**
	 * Mthode pour dterminer si deux nombres de type double sont <b>relativement gaux</b>. 
	 * En utilisant une approche de calcul de diffrence, on vrifie si
	 * <ul>a - b < EPSILON*ref</ul>  
	 * afin de <b>valid l'galit</b> entre a et b (a == b). EPSILON est un seuil de prcision 
	 * et ref est une base de rfrence (la valeur absolue la plus leve parmis a et b).
	 * <p>Cenpendant, si les deux chiffres sont infrieurs  EPSILON, ils seront considrs comme gaux.</p>
	 * 
	 * @param a Le 1ier nombre  comparer.
	 * @param b Le 2ime nombre  comparer.
	 * @param epsilon - La prcision acceptable.
	 * @return <b>true</b> si les deux nombres sont <b>relativement gaux</b> et <b>false</b> sinon.
	 */
	public static boolean nearlyEquals(double a, double b, double epsilon)
	{
	  double absA = Math.abs(a);
    double absB = Math.abs(b);
    double diff = Math.abs(a - b);
    
    // Vrification du cas particulier : 0 = 0 et infiny = infiny (mais pas certain ...) 
    if(a == b)                            
      return true;
     
    // Cas des petites chiffres : Vrifier si les deux chiffres sont trs prs l'un de l'autre
    if(diff < epsilon)  
      return true;
      
    // Cas gnral
    double positive_max;
        
    if(absA > absB)
      positive_max = absA;
    else
      positive_max = absB;
      
    if(diff < positive_max*epsilon)
      return true;
    else
      return false;
	}
	
	/**
	 * Mthpde pour dterminer si une valeur est relativement prs de zro.
	 * Cette mthode est ncessaire, car une opration mathmatique menant au chiffre 0 peut tre 0.0 et -0.0 ce qui n'est pas gal selon JAVA.
	 * 
	 * @param value La valeur  comparer avec 0.
	 * @return <b>true</b> si la valeur est <b>relativement gal</b>  zro et <b>false</b> sinon.
	 */
	public static boolean nearlyZero(double value)
	{
	  return nearlyEquals(value, 0.0);
	}
		
	/**
	 * Mthode permettant d'valuer la racine relle d'un polynme de degr '1' de la forme
	 * <ul>Ax + B = 0.</ul>
	 * <p>Un polynme de degr '1' possde au maximum <b>une</b> racine relle.</p>
	 *  
	 * @param A Le coefficient devant le terme de puissance '1' (x).
	 * @param B Le coefficient devant le terme de puissance '0' (1).
	 * @return La racine relle d'un polynme de degr '1' (s'il y en a).
	 * @throws SInfinityOfSolutionsException Si le polynme contient une infinit de solution (0x + 0 = 0).
	 */
	public static double[] linearRealRoot(double A, double B) throws SInfinityOfSolutionsException
	{
	  // Vrifier si le polynme n'est pas d'un degr infrieur.
	  if(nearlyZero(A))
	  {
	    // Vrifier si l'quation n'est pas 0*t + B = 0.
	    // Si B == 0, l'quation admettra une infinit de solutions.
	    // Autrement, l'quation n'admettra aucune solution.
	    if(nearlyZero(B))
	      throw new SInfinityOfSolutionsException("L'quation linaire 0x + 0 = 0 admet une infinit de solutions.");
	    
	    return NO_ROOT_SOLUTION;
	  }
	  
	//----------------------------------------------------------------------------------------------
    // VERSION LABORATOIRE :
    // Remplacez cette exception par votre implmentation.
    //----------------------------------------------------------------------------------------------
	
	 throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
	
	}
	
	/**
   * Mthode permettant d'valuer les racines relles d'un polynme de degr '2' de la forme
   * <ul>Ax^2 + Bx + C = 0.</ul>
   * <p>Un polynme de degr '2' possde au maximum <b>deux</b> racines relles.</p>
   *  
   * @param A Le coefficient devant le terme de puissance '2' (x^2).
   * @param B Le coefficient devant le terme de puissance '1' (x).
   * @param C Le coefficient devant le terme de puissance '0' (1).
   * @return Les racines relles de d'un polynme de degr '2' (s'il y en a). Les solutions retournes dans un tableau sont <b>tries en ordre croissant</b>. 
   * @throws SInfinityOfSolutionsException Si le polynme contient une infinit de solution (0x^2 + 0x + 0 = 0).
   */
	public static double[] quadricRealRoot(double A, double B, double C) throws SInfinityOfSolutionsException
	{
	  // Vrifier si le polynme n'est pas d'un degr infrieur
	  if(nearlyZero(A))
	    return linearRealRoot(B,C);
	  	  
	  //----------------------------------------------------------------------------------------------
	  // VERSION LABORATOIRE :
	  // Remplacez cette exception par votre implmentation.
	  //----------------------------------------------------------------------------------------------
	  throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
	  
	}
	
	/**
   * Mthode permettant d'valuer les racines relles d'un polynme de degr '3' de la forme
   * <ul>Ax^3 + Bx^2 + Cx + D = 0.</ul>
   * <p>Un polynme de degr '3' possde au maximum <b>trois</b> racines relles.</p>
   *  
   * @param A Le coefficient devant le terme de puissance '3' (x^3).
   * @param B Le coefficient devant le terme de puissance '2' (x^2).
   * @param C Le coefficient devant le terme de puissance '1' (x).
   * @param D Le coefficient devant le terme de puissance '0' (1).
   * @return Les racines relles de d'un polynme de degr '3' (s'il y en a). Les solutions retournes dans un tableau sont <b>tries en ordre croissant</b>. 
   * @throws SInfinityOfSolutionsException Si le polynme contient une infinit de solution (0x^3 + 0x^2 + 0x + 0 = 0).
   */
  public static double[] cubicRealRoot(double A, double B, double C, double D) throws SInfinityOfSolutionsException
  {
    // Vrifier si le polynme n'est pas d'un degr infrieur
    if(nearlyZero(A))
      return quadricRealRoot(B,C,D);
    
    throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
    
  }
  
	/**
   * Mthode permettant d'valuer les racines relles d'un polynme de degr '4' de la forme
   * <ul>Ax^4 + Bx^3 + Cx^2 + Dx + E = 0.</ul>
   * <p>Un polynme de degr '4' possde au maximum <b>quatre</b> racines relles.</p>
   *  
   * @param A Le coefficient devant le terme de puissance '4' (x^4).
   * @param B Le coefficient devant le terme de puissance '3' (x^3).
   * @param C Le coefficient devant le terme de puissance '2' (x^2).
   * @param D Le coefficient devant le terme de puissance '1' (x).
   * @param E Le coefficient devant le terme de puissance '0' (1).
   * @return Les racines relles de d'un polynme de degr '4' (s'il y en a). Les solutions retournes dans un tableau sont <b>tries en ordre croissant</b>. 
   * @throws SInfinityOfSolutionsException Si le polynme contient une infinit de solution (0x^4 + 0x^3 + 0x^2 + 0x + 0 = 0).
   */
	public static double[] quarticRealRoot(double A, double B, double C, double D, double E) throws SInfinityOfSolutionsException
	{
	  // Vrifier si le polynme n'est pas d'un degr infrieur.
	  if(nearlyZero(A))
	    return cubicRealRoot(B,C,D,E);
	  
	  throw new SNoImplementationException("Erreur SMath : C'est mthode n'a pas t implmente.");
    
	}
	
	/**
   * Mthode qui effectue le calcul inverse de l'interpolation linaire d'une valeur numrique.
   * Cela sigifie que l'on cherche la valeur du paramtre d'interpolation t  partir d'une valeur interpole
   * ainsi que des deux valeurs extrmes.
   * 
   * @param v La valeur interpole dont la valeur de t doit tre calcule.
   * @param v0 La valeur de rfrence pondre par 1 - t.
   * @param v1 La valeur de rfrence pondre par le facteur t.
   * @return La facteur t d'interpolation linaire.
   */
  public static double reverseLinearInterpolation(double v, double v0, double v1) 
  {
    //  partir de la relation v = (1-t)*v0 + t*v1 , on doit isoler t.
    return (v - v0)/(v1 - v0);
  }
  
  /**
   * Mthode qui effectue le calcul de l'interpolation linaire d'une valeur numrique.
   * 
   * @param v0 La valeur de rfrence pondre par 1 - t.
   * @param v1 La valeur de rfrence pondre par le facteur t.
   * @param t Le paramtre de pondration.
   * @return La valeur interpole.
   */
  public static double linearInterpolation(double v0, double v1, double t) 
  {   
    // Calcul de l'interpolation : v = v0*(1 - t) + v1*t
    return v0*(1.0 - t) + (v1*t);
  }
  
  /**
   * Mthode qui effectue le calcul de l'interpolation quadratique d'une valeur numrique.
   * 
   * @param v0 La valeur de rfrence pondre par 1 - t*t.
   * @param v1 La valeur de rfrence pondre par le facteur t*t.
   * @param t Le paramtre de pondration.
   * @return La valeur interpole.
   */
  public static double quadricInterpolation(double v0, double v1, double t) 
  {
    // Calcul de l'interpolation : v = v0*(1 - t^2) + v1*t^2
    double t2 = t*t;
    
    return v0*(1.0 - t2) + v1*t2;
  }
  
  /**
   * Mthode dterminant le signe d'un nombre. Les rsultats sont
   * <ul> -1 si a < 0</ul>
   * <ul> 0 si a = 0</ul>
   * <ul> 1 si a > 0</ul>
   * 
   * @param a Le nombre.
   * @return Le signe du nombre.
   */
  public static double sgn(double a)
  {
    if(a > 0)
      return 1.0;
    else
      if(a < 0)
        return -1.0;
      else
        return 0.0;
  }
  
  /**
   * Mthode permettant de gnrer un nombre alatoire entre une valeur minimale et maximale.
   *  
   * @param min La valeur minimale du nombre alatoire.
   * @param max La valeur maximale du nombre alatoire.
   * @return Le nombre alatoire.
   */
  public static double random(double min, double max)
  {
    return Math.random()*(max - min) + min;
  }
  
  /**
   * Mthode permettant de gnrer un nombre alatoire entre une valeur minimale et maximale.
   * 
   * @param min La valeur minimale du nombre alatoire.
   * @param max La valeur maximale du nombre alatoire.
   * @return Le nombre alatoire.
   */
  public static int random(int min, int max)
  {
    return (int)Math.round((Math.random()*(max - min))) + min;
  }
  /**
   * Mthode permettant de gnrer alatoirement une valeur vrai ou faux.
   * 
   * @return alatoirement une rponse <b>true</b> ou <b>false</b>.
   */
  public static boolean randomTrueOrFalse()
  {
    return Math.random() > 0.5;
  }
  
  /**
   * Mthode pour gnrer un nombre alatoire appartenant  un tableau de valeurs.
   * Les choix des lments dans le tableau sont tous quiprobable.
   * 
   * @param possibility Le tableau des valeurs possibles.
   * @return Une valeur alatoire appartenant au tableau des possibilits. 
   */
  public static int random(int[] possibility)
  {
    // Choisir un indice du tableau alatoirement.
    int random_index = (int)(Math.random()*possibility.length);
    
    return possibility[random_index];
  }
  
  /**
   * Mthode pour gnrer un nombre alatoire appartenant  un tableau de valeurs dont le poids probabiliste
   * n'est pas quiprobable. La probabilit de chaque valeur dpend d'un poid dtermin pour chaque valeur.
   * L'association est effectue selon les indices des deux tableaux. 
   * 
   * @param possibility Le tableau des valeurs possibles.
   * @param weight Le tableau des poids du choix alatoire.
   * @return Une valeur alatoire appartenant au tableau des possibilits. 
   * @throws SIllegalNegativeValueException Si le tableau des poids contient une valeur ngative.
   */
  public static int random(int[] possibility, double[] weight) throws SIllegalNegativeValueException
  {
    // Vrification de la longueur des tableaux.    
    if(possibility.length != weight.length)
      throw new SRuntimeException("Le tableau possibility possde une longueur " + possibility.length  + " et le tableau weight possde une longueur " + weight.length + " ce qui n'est pas gal.");
    
    // valuer la somme des pondratoire.
    double sum = 0.0;
    
    for(double p : weight)
      if(p < 0.0)
        throw new SIllegalNegativeValueException("Erreur SMath 005 : La valeur ngative " + p + " n'est pas adquate dans le tableau des poids.");
      else
        sum += p;
    
    // Gnrer un nombre alatoire entre 0 et la pondratoire total.
    double random = Math.random()*sum;
    
    // Choisir l'tat alatoire pondr par le poids des choix.
    int index = 0;
    
    while(random > weight[index])
    {
      random -= weight[index];
      index++;
    }
    
    // Retourner le choix alatoire pondr par le poids des choix.  
    return possibility[index];
  }
  
}
