/**
 * 
 */
package sim.parser.model.obj;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import sim.exception.SConstructorException;
import sim.util.SBufferedReader;
import sim.util.SFileSearch;
import sim.util.SLog;
import sim.util.SStringUtil;

/**
 * Classe reprsentant un parser de modle 3d de format obj (Wavefront).
 * @author Simon Vzina
 * @since 2015-03-19
 * @version 2015-08-03
 */
public class SModelOBJParser {

  private static final String MODEL_EXTENSION = "obj";
  
  private static final int VERTEX_CODE = 1;
  private static final int TEXTURE_CODE = 2;
  private static final int NORMAL_CODE = 3;
  private static final int FACE_CODE = 4;
  private static final int MTL_LIBRARY_CODE = 5;
  private static final int USEMTL_CODE = 6;
  
  private static final Hashtable<String,Integer> keyword_table = initialiseKeywordTable();
  
  private final String file_name;                 //nom du fichier .obj
  
  private final ArrayList<SVertex> vertex_list;   //liste des vertex du modle 3d(commenant  l'index 1)
  
  private final ArrayList<SVertex> texture_list;  //liste des vertex du modle 3d(commenant  l'index 1)
  
  private final ArrayList<SVertex> normal_list;   //liste des normale du modle 3d(commenant  l'index 1)
  
  private final List<SFace> face_list;		        //liste des faces (index des vertex pour dcrire les formes gomtriques)
  
  private final List<SPolygonOBJ> polygon_list;   //liste des polygones du modle
  
  private String current_material_name;           //nom du matriel prsentement  affecter au Face en lecture.
  
  private final List<SMaterialOBJ> material_list; //liste des matriels
  
  /**
   * Constructeur d'un parser pour objet 3d de de format obj.
   * @param file_name - Le nom du fichier comportant le modle 3d de format obj.
   * @throws SModelOBJParserException Si le fichier n'est pas adquat. 
   */
  public SModelOBJParser(String file_name)throws SModelOBJParserException
  {
    this.file_name = file_name;
    
    vertex_list = new ArrayList<SVertex>();
    vertex_list.add(null);  //l'index 0 n'est pas admissible      
    
    texture_list = new ArrayList<SVertex>();
    texture_list.add(null);  //l'index 0 n'est pas admissible
    
    normal_list = new ArrayList<SVertex>();
    normal_list.add(null);  //l'index 0 n'est pas admissible
    
    current_material_name = SMaterialOBJ.DEFAULT_NAME;  //nom par dfaut du matriel  affect  une face
    face_list = new LinkedList<SFace>();
        
    material_list = new LinkedList<SMaterialOBJ>();
    
    parseOBJ(file_name);    //lecture du fichier et construction des trois types de vertex et des faces
    
    polygon_list = new LinkedList<SPolygonOBJ>();
    buildPolygon();         //construction des polygones en affectant les vertex selon la description des faces
  
    //Vrification si le fichier OBJ ne contient pas de rfrence  un fichier MTL.
    //Si c'est le cas, il faut construire un SMaterialOBJ par dfaut afin que la 
    //rfrence au nom du matriel "default" puisse exister 
    if(material_list.isEmpty())
    {
      SLog.logWriteLine("Message SModelOBJParser : Le modle prsent ne possde pas de matriel. Un matriel par dfaut sera gnr.");
      material_list.add(new SMaterialOBJ());
    }
  }

  /**
   * Mthode pour obtenir le nom du fichier du modle 3d de format OBJ.
   * @return Le nom du fichier.
   */
  public String getFileName()
  {
	  return file_name;
  }
  
  /**
   * Mthode pour obtenir la liste des polygones dfinissant le modle 3d de format OBJ.
   * @return La liste des polygones du modle.
   */
  public List<SPolygonOBJ> getListPolygon()
  {
	  return polygon_list;
  }
  
  /**
   * Mthode pour obtenir la liste des matriels du modle 3d de format OBJ. Dans cette liste,
   * on y retrouve un matriel par dfaut (portant le nom "default"). Ce matriel peut se retrouver
   * plusieurs fois dans la liste, car il sera prsent pour chaque fichier MTL lu par le modle.
   * @return La liste des matriaux du modle.
   */
  public List<SMaterialOBJ> getListMaterial()
  {
    return material_list;
  }
  
  /**
   * Mthode pour faire la table de correspondance entre un mot cl un un code d'action.
   * @return La table de mot cl.
   */
  private static Hashtable<String,Integer> initialiseKeywordTable()
  {
    Hashtable<String,Integer> table = new Hashtable<String,Integer>();
    
    table.put("v", new Integer(VERTEX_CODE));
    table.put("vt", new Integer(TEXTURE_CODE));
    table.put("vn", new Integer(NORMAL_CODE));
    table.put("f", new Integer(FACE_CODE));
    table.put("mtllib", new Integer(MTL_LIBRARY_CODE));
    table.put("usemtl", new Integer(USEMTL_CODE));
    
    return table;
  }
  
  /**
   * Mhode pour faire la lecture d'un fichier obj  partir du nom du fichier.
   * @param file_name - Le nom du fichier obj  lire.
   * @throws SModelOBJParserException Si le fichier n'est pas dans le bon format.
   * @throws SModelOBJParserException Si le fichier n'est pas trouv.
   * @throws SModelOBJParserException Si le fichier existe sous plusieurs versions dans les diffrents sous-rpertoires.
   */
  private void parseOBJ(String file_name)throws SModelOBJParserException
  {
    //Vrification du format du fichier
    if(!SStringUtil.extensionFileLowerCase(file_name).equals(MODEL_EXTENSION))
    	throw new SModelOBJParserException("Erreur SModelOBJParser 001 : Le fichier '" + file_name + "' n'est pas dans le format '" + MODEL_EXTENSION + "'.");
    
    //Trouver le fichier  partir du rpertoire o l'excution de l'application est ralise
    SFileSearch search = new SFileSearch("", file_name);
    
    if(!search.isFileFound())
      throw new SModelOBJParserException("Erreur SModelOBJParser 002 : Le fichier '" + file_name + "' n'est pas trouv.");
    
    if(search.isManyFileFound())
      throw new SModelOBJParserException("Erreur SModelOBJParser 003 : Le fichier '" + file_name + "' a t trouv plus d'une fois dans les diffrents sous-rpertoires. Veuillez en garder qu'une seule version.");
    
    File file = new File(search.getFileFoundList().get(0)); //prendre la 1ier trouv dans la liste
    
    try
    {
      FileReader fr = new FileReader(file);
      SBufferedReader sbr = new SBufferedReader(fr);
      parseOBJ(sbr);
      
    }catch(FileNotFoundException e){
      throw new SModelOBJParserException("Erreur SModelOBJParser 004 : IMPOSSIBLE! La localisation du fichier a t dj test.");     
    }catch(IOException ioe){
      throw new SModelOBJParserException("Erreur SModelOBJParser 005 : Une erreur de type I/O est survenue.");      
    }
  }
  
  /**
   * Mthode pour lire le fichier de format obj  l'aide d'un SBufferedReader.
   * @param sbr - Le buffer de lecture.
   * @throws IOException S'il y a eu une erreur de type I/O lors de la lecture du fichier.
   * @throws SModelOBJParserException S'il y a eu une erreur menant  une incapacit de dfinir le modle 3d.
   */
  private void parseOBJ(SBufferedReader sbr)throws SModelOBJParserException, IOException
  {			          
	  boolean end_reading = false;	      //condition de fin de lecture
	  
	  while(!end_reading)
	  {
	    String line = sbr.readLine();    //ligne courante en lecture
		  
	    //Si la ligne lu est "null", la fin du fichier est atteinte
		  if(line == null)
			  end_reading = true;
		  else
		  {
			  StringTokenizer tokens = new StringTokenizer(line);  //fragmenter la ligne lue en liminant les caractres d'espacement
			  
			  //S'il y a un token  analyser
			  if(tokens.hasMoreTokens())
			  {
			    String keyword = tokens.nextToken();     //obtenir le mot cl tant le 1ier mot de la ligne           
			  			  
			    //Si le mot cl doit tre analys
  			  if(keyword_table.containsKey(keyword))
  			  {
    				int code = (keyword_table.get(keyword)).intValue();       //code de lecture
    				
    				//Obtenir la ligne restante (sans 1ier caractre d'espacement + mot cl + autres caractres d'espacement + ligne restante ...)
    				String remaining_line = SStringUtil.removeAllFirstSpacerCaracter(line);
    				remaining_line = remaining_line.substring(keyword.length());	
    				remaining_line = SStringUtil.removeAllFirstSpacerCaracter(remaining_line);
    				    				
    				try{
    				  parseOBJLine(code, remaining_line);                     //analyse de la ligne restante
    				}catch(SModelOBJParserException e){
    				  throw new SModelOBJParserException("Ligne " + (sbr.atLine()-1) + " - " + e.getMessage());
    				}
  			  }
			  }  
		  }
	  }//fin while end_reading
  }
  
  /**
   * Mthode pour faire l'analyse de l'expression d'un restant de ligne en fonction du code de lecture lu 
   * au dbut de la ligne.
   * @param code - Le code reprsentant la dfinition de l'expression restant  analyser.
   * @param remaining_line - L'expression  analyser sera le code de lecture.
   * @throws SModelOBJParserException S'il y a eu une erreur de lecture menant  un chec de la lecture du modle 3d.
   */
  private void parseOBJLine(int code, String remaining_line)throws SModelOBJParserException
  {
	  switch(code)
	  {
	  	case VERTEX_CODE :  try{
	  	                    vertex_list.add(new SVertex(remaining_line));
	  	                    }catch(SConstructorException e){
	  	                      throw new SModelOBJParserException("Erreur SModelOBJParser 006 : Un vertex est mal dfini. " + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
	  	                    }
	  	                    break;
	  	
	  	case TEXTURE_CODE : try{
	  	                    texture_list.add(new SVertex(remaining_line)); 
	  	                    }catch(SConstructorException e){
                            throw new SModelOBJParserException("Erreur SModelOBJParser 007 : Un vertex texture est mal dfini." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
                          }
	  	                    break;
	  	
	  	case NORMAL_CODE :  try{
	  	                    normal_list.add(new SVertex(remaining_line));
	  	                    }catch(SConstructorException e){
                            throw new SModelOBJParserException("Erreur SModelOBJParser 008 : Un vertex normal est mal dfini." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
                          }
	  	                    break;
	  	                          
	  	case FACE_CODE :    try{
	  	                    face_list.add(new SFace(current_material_name, remaining_line));
	  	                    }catch(SConstructorException e){
	  	                      SLog.logWriteLine("Message SModelOBJParser : Une face est mal dfinie." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
	  	                    }
	  	                    break;
	  	
	  	case USEMTL_CODE : current_material_name = remaining_line; 
	  	                   break;
	  	
	  	case MTL_LIBRARY_CODE : try{
	  	                        
	  	                        SLog.logWriteLine("Message SModelOBJParser : Lecture de la librairie de matriaux '" + remaining_line + "'.");
	  	                        
	  	                        SMaterialOBJParser p = new SMaterialOBJParser(remaining_line);  //lire la librairie des matriels
	  	                        List<SMaterialOBJ> l = p.getMaterialList();                     //obtenir la liste des matriels
	  	                        material_list.addAll(l);                                        //ajouter l'ensemble des matriels  la liste du matriel
	  	                        }catch(SMaterialOBJParserException e){
	  	                          SLog.logWriteLine("Message SModelOBJParser : Une librairie de matriaux est mal dfinie." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage());
	  	                        }
	  	                        break;
	  }
  }
  
  /**
   * Mthode pour construire l'ensemble des polygones du modle  partir de la liste des faces (descriptions des vertex requis identifis par indexage dans les listes).
   */
  private void buildPolygon()
  {
    //Construire l'ensemble des polygones du modle
    for(SFace f : face_list)
      polygon_list.add(new SPolygonOBJ(f, vertex_list, texture_list, normal_list));
  }
     
  @Override
  public String toString()
  {
    String vertex_nb = "Vertex number : " + (vertex_list.size()-1);
    String texture_nb = "Texture number : " + (texture_list.size()-1);
    String normal_nb = "Normal number : " + (normal_list.size()-1);
    String face_nb = "Face number :   " + face_list.size();
    String material_nb = "Material number :   " + material_list.size();
    
    return vertex_nb + SStringUtil.END_LINE_CARACTER + texture_nb + SStringUtil.END_LINE_CARACTER + normal_nb + SStringUtil.END_LINE_CARACTER + face_nb + SStringUtil.END_LINE_CARACTER + material_nb + SStringUtil.END_LINE_CARACTER;
  }
  
  /**
   * Mthode de test pour la classe SFace.
   * @param args
   */
  public static void main(String[] args) throws FileNotFoundException, SModelOBJParserException
  {
    test1();
    test2();
  }
  
  /**
   * Mthode de test #1. 
   */
  private static void test1() throws FileNotFoundException, SModelOBJParserException
  {
    String model_name = "lamp.obj";
    SModelOBJParser parser = new SModelOBJParser(model_name);
    
    System.out.println(model_name);
    System.out.println(parser);
  }
 
  /**
   * Mthode de test #1. 
   */
  private static void test2() throws FileNotFoundException, SModelOBJParserException
  {
    String model_name = "foot.obj";
    SModelOBJParser parser = new SModelOBJParser(model_name);
        
    System.out.println(model_name);
    System.out.println(parser);
  }
}//fin SModelOBJParser
