/**
 * 
 */
package sim.physics;

import sim.exception.SConstructorException;
import sim.math.SVector3d;

/**
 * La classe <b>SWave</b> reprsente une onde physique gnr par un oscillateur harmonique simple.
 * La forme de l'onde sera alors une onde sinusodale dont la frquence et la phase initiale seront dtermines par son oscillateur.
 * La longueur d'onde l'onde sera dtermine par la vitesse de propagation du milieu et la frquence de l'oscillateur. 
 * Dans le cas d'une onde lectromagntique voyageant dans le vide, la vitesse de propagation sera gale  <i>c</i>.
 * 
 * @author Simon Vzina
 * @since 2016-02-22
 * @version 2016-02-23
 */
public class SWave {

  /**
   * La variable <b>oscillator</b> correspond  l'oscillateur qui gnre l'onde .
   */
  private final SOscillator oscillator;
  /**
   * La variable <b>wave_speed</b> correspond  la vitesse de propagation des ondes dans le milieu par dfaut.
   */
  private final double wave_speed;
    
  /**
   * La variable <b>position</b> correspond  la position o l'onde doit tre value.
   */
  private SVector3d position;
  
  //--------------------------
  // Paramtre de prcalcul //
  //--------------------------
  
  /**
   * La variable <b>k</b> correspond au nombre d'onde de l'onde.
   */
  private final double k;
  
  /**
   * La variable <b>kx</b> correspond  la phase spatiale de l'onde  l'endroit dsign par la variable <b>position</b>.
   */
  private double kx;
  
  /**
   * La variable <b>A</b> correspond  l'amplitude de l'onde  l'endroit dsign par la variable <b>position</b>.
   */
  private double A;
  
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * Constructeur d'une onde lectromagntique progressive sinusodale voyageant dans le vitesse  la vitesse de la lumire.
   * 
   * @param oscillator L'oscillateur harmonique simple  l'origine de l'onde progressive.
   */
  public SWave(SOscillator oscillator)
  {
    this(oscillator, SPhysics.c);
  }
  
  /**
   * Constructeur d'une onde progressive sinusodale voyageant dans un milieu homogne  une vitesse constante.
   * 
   * @param oscillator L'oscillateur harmonique simple  l'origine de l'onde progressive.
   * @param wave_speed La vitesse de propagation de l'onde dans le milieu.
   * @throws SConstructorException Si la vitesse de l'onde est ngative.
   */
  public SWave(SOscillator oscillator, double wave_speed) throws SConstructorException
  {
    if(wave_speed <= 0)
      throw new SConstructorException("Erreur SWave 001 : La vitesse de l'onde '" + wave_speed + "' doit tre positive et non nulle.");
  
    this.oscillator = oscillator;
    
    this.wave_speed = wave_speed;
    
    k = evaluateWaveNumber(oscillator.getFrequency(), wave_speed);
    
    setPosition(oscillator.getPosition());
  }

  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir l'oscillateur qui gnre l'onde progressive.
   * 
   * @return L'oscillateur.
   */
  public SOscillator getOscillator()
  {
    return oscillator;
  }
  
  /**
   * Mthode pour obtenir la vitesse de propagation de l'onde progressive.
   * 
   * @return La vitesse de l'onde.
   */
  public double getWaveSpeed()
  {
    return wave_speed;
  }
  
  /**
   * Mthode pour obtenir la phase spatiale de l'onde (paramtre k*x de l'onde progressive sinusodale)
   *  l'endroit dtermin par la mthode setPosition(). Si la prcdente mthode n'a pas t appele avant,
   * la position o sera value la phase spatiale sera  l'endroit o l'oscillateur est situ ce qui engendra un phase nulle.
   *  
   * @return La phase spatiale  une endroit pralablement dtermin. 
   */
  public double getSpacePhase()
  {
    return kx;
  }
  
  /**
   * Mthode pour obtenir la phase spatiale de l'onde (paramtre k*x de l'onde progressive sinusodale)
   *  l'endroit dsign par le point p.
   * 
   * @param p Le point o sera valu la phase spatiale de l'onde.
   * @return La phase spatiale de l'onde au point p.
   */
  public double getSpacePhase(SVector3d p)
  {
    return evaluateSpacePhase(p);
  }
  
  /**
   * <p>
   * Mthode pour dfinir la position o l'onde sera value. Cette position porte galement le nom de point P.
   * </p>
   * <p>
   * En fixant cette position, ceci permet de prcalculer la phase spatiale de l'onde correspond au paramtre <b>k*x</b> de l'onde dans l'expression
   * <ul>x = A sin (k*x  omega*t + phi)</ul>
   * </p>
   * 
   * @param p La position du point P.
   */
  public void setPosition(SVector3d p)
  {
    position = p;
    
    kx = evaluateSpacePhase(position);
    
    // POUR L'INSTANT ... IL N'Y A PAS DE PARAMTRE D'ATTNUATION CAUSE PAR LA PROPAGATION EN 3D DE L'ONDE !!!
    A = oscillator.getAmplitude();
  }
  
  /**
   * Mthode pour obtenir la valeur de l'onde progressive au temps t  partir de l'expression
   * <ul>y = A*sin(k*x - omega*t + phi)</ul>
   * 
   * @param t Le temps.
   * @return La valeur de l'onde progressive.
   */
  public double getValue(double t)
  {
    return A*Math.sin(kx - oscillator.getTimePhase(t) + oscillator.getInitialPhase());
  }
  
  /**
   * <p>
   * Mthode permettant d'valuer le nombre d'onde de l'onde.
   * </p>
   * <p>
   *  partir de la dfinition k = 2*Pi/lamda et lamda = v/f, nous obtenons
   * <ul>k = 2*Pi*f/v</ul>
   * </p>
   * 
   * @param f La frquence de l'oscillateur.
   * @param v La vitesse de propagation de l'onde.
   * @return Le nombre d'onde de l'onde.
   */
  private double evaluateWaveNumber(double f, double v)
  {
    return 2*Math.PI*f/v;
  }
  
  /**
   * Mthode permettant d'valuer la phase spatiale  partir d'une position p o l'on dsire l'value.
   * Ceci correspond au paramtre <b>k*x</b> de l'onde dans l'expression
   * <ul>x = A sin (k*x  omega*t + phi)</ul>
   *  
   * @param p Le point p.
   * @return La phase spatiale.
   */
  private double evaluateSpacePhase(SVector3d p)
  {
    // Dfinir la position x o l'on dsire valuer l'onde
    double x = p.substract(oscillator.getPosition()).modulus();
    
    // Dfinir le produit kx entre 0 et 2 Pi
    return SWaveOptics.phaseBetweenZeroAnd2Pi(k*x);
  }
  
}//fin de la classe SWave
