/**
 * 
 */
package sim.util;

import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

/**
 * La classe <b>SBrowser</b> reprsente un explorateur de rpertoire et de fichier.
 * 
 * @author Simon Vzina
 * @since 2018-02-26
 * @version 2019-12-01
 */
public class SBrowser {

  /**
   * La constante <b>DEFAULT_MAX_DEPTH</b> reprsente le niveau de profondeur maximal de recherche rcursive dans les sous-rpertoire.
   */
  private final int DEFAULT_MAX_DEPTH = 8;
  
  /**
   * La variable <b>file</b> reprsente l'emplacement courant de l'explorateur.
   */
  private File file;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un explorateur de rpertoire.
   */
  public SBrowser()
  {
    this.file = new File(System.getProperty("user.dir"));
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir le nom du rpertoire o est situ l'explorateur de rpertoire.
   * 
   * @return
   */
  public String getDirectoryName()
  {
    return this.file.getName();
  }
  
  /**
   * Mthode pour obtenir le chemin o est situ l'explorateur de rpertoire.
   * 
   * @return Le chemin de localisation de l'explorateur.
   */
  public String getPath()
  {
    return this.file.getPath();
  }
  
  /**
   * Mthode qui retourne une liste d'accs au fichier visible par la localisation courant de l'explorateur.
   * 
   * @return Un liste d'accs aux fichiers.
   */
  public List<File> getFileList()
  {
    List<File> list = new ArrayList<File>();
    
    for(File f : file.listFiles())
      if(f.isFile())
        list.add(f);
    
    return list;
  }
  
  /**
   * Mthode qui retourne une liste d'accs au rpertoire visible par la localisation courant de l'explorateur.
   * 
   * @return Un liste d'accs aux rpertoires.
   */
  public List<File> getDirectoryList()
  {
    List<File> list = new ArrayList<File>();
    
    for(File f : this.file.listFiles())
      if(f.isDirectory())
        list.add(f);
    
    return list;
  }
  
  /**
   * Mthode pour dteminer si le browser est situ dans un rpertoire vide.
   * 
   * @return <b>true</b> si le rpertoire est vide et <b>false</b> sinon.
   */
  public boolean isEmpty()
  {
    return this.file.listFiles().length == 0;
  }
  
  /**
   * Mthode pour entrer dans un rpertoire  partir de la position local du browser.
   * 
   * @param directory_name Le nom du rpertoire dans lequel on veut entrer.
   * @throws FileNotFoundException Si le nom du rpertoire n'a pas t trouv.
   * @throws SManyFilesFoundException Si plusieurs rpertoires ayant le mme nom a t trouvs.
   */
  public void forward(String directory_name) throws FileNotFoundException, SManyFilesFoundException
  {
    List<File> list = getDirectoryList();
    List<File> found_list = new ArrayList<File>();
    
    for(File f : list)
      if(f.getName().equalsIgnoreCase(directory_name))
        found_list.add(f);
    
    if(found_list.isEmpty())
      throw new FileNotFoundException("Erreur SBrowser 001 : Le rpertoire '" + directory_name + "' n'a pas t trouv  partir de l'emplacement initiale '" + getPath() + "'.");
    
    if(found_list.size() > 1)
      throw new SManyFilesFoundException("Erreur SBrowser 002 : Le rpertoire '" + directory_name + "' a t trouv  plusieurs reprise  partir de l'emplacement initiale '" + getPath() + "'. Voici la liste : " + found_list);
    
    // Affectation de la nouvelle position de l'explorateur.
    this.file = found_list.get(0);
  }
  
  /**
   * Mthode pour sortir d'un rpertoire.
   */
  public void back()
  {
    this.file = new File(file.getParent());
  }
  
  /**
   * Mthode pour construire un fichier  partir de l'emplacement du browser.
   * 
   * @param file_name Le nom du fichier.
   * @param extension Le nom de l'extension du fichier.
   * @return Le fichier.
   */
  public File createFile(String file_name, String extension)
  {
    // Gnrer le nom du fichier avec le bon sous-rpertoire.
    String text_name = this.getPath() + SStringUtil.FILE_SEPARATOR_CARACTER + file_name + "." + extension;
   
    return new File(text_name);
  }
  
  /**
   * Mthode pour construire un rpertoire  partir de l'emplacement du browser.
   * 
   * @param directory_name Le nom du rpertoire  crer.
   * @return <b>true</b> si le rpertoire a t cr et <b>false</b> sinon.
   */
  public boolean createDirectory(String directory_name)
  {
    File new_directory = new File(this.file.getPath() + SStringUtil.FILE_SEPARATOR_CARACTER + directory_name);
    
    return new_directory.mkdirs();
  }
  
  /**
   * Mthode pour effacer un rpertoire ainsi que l'ensemble de son contenu (rpertoire et fichier).
   * Le browser doit tre situ  l'emplacement du rpertoire  effacer.
   * Il n'y aura pas de recherche rcursive du nom du rpertoire  effacer.
   * 
   * @param directory_name Le nom du rpertoire  effacer (ainsi que son contenu).
   */
  public void deleteDirectory(String directory_name)
  {
    File directory = new File(this.file.getPath() + SStringUtil.FILE_SEPARATOR_CARACTER + directory_name);
    
    deleteDirectory(directory);
  }
   
  /**
   * Mthode pour dplacer l'explorateur  l'emplacement d'un rpertoire  partir du rpertoire courant de l'explorateur.
   * 
   * @param directory_name Le nom du rpertoire vers lequel l'explorateur se dirigera.
   * @throws FileNotFoundException Si le nom du rpertoire n'a pas t trouv.
   * @throws SManyFilesFoundException Si plusieurs rpertoires ayant le mme nom a t trouvs.
   */
  public void goTo(String directory_name) throws FileNotFoundException, SManyFilesFoundException
  {
    goTo(directory_name, DEFAULT_MAX_DEPTH);
  }
  
  /**
   * Mthode pour dplacer l'explorateur  l'emplacement d'un rpertoire  partir du rpertoire courant de l'explorateur.
   * 
   * @param directory_name Le nom du rpertoire vers lequel l'explorateur se dirigera.
   * @throws FileNotFoundException Si le nom du rpertoire n'a pas t trouv.
   * @throws SManyFilesFoundException Si plusieurs rpertoires ayant le mme nom a t trouvs.
   */
  public void goTo(String directory_name, int max_depth) throws FileNotFoundException, SManyFilesFoundException
  {
    List<File> list = new ArrayList<File>();
    
    searchDirectories(directory_name, file, 0, max_depth, list);
    
    if(list.isEmpty())
      throw new FileNotFoundException("Erreur SBrowser 003 : Le rpertoire '" + directory_name + "' n'a pas t trouv  partir de l'emplacement initiale '" + getPath() + "'.");
    
    if(list.size() > 1)
      throw new SManyFilesFoundException("Erreur SBrowser 004 : Le rpertoire '" + directory_name + "' a t trouv  plusieurs reprise  partir de l'emplacement initiale '" + getPath() + "'. Voici la liste : " + list);
    
    // Affectation de la nouvelle position de l'explorateur.
    this.file = list.get(0);
  }
  
  /**
   * Mthode pour obtenir une liste d'accs  un fichier correspondant  un nom particulier.
   * 
   * @param file_name Le nom du fichier en recherche (avec extension).
   * @return Une liste de fichiers ayant le nom dsir.
   */
  public List<File> searchFiles(String file_name)
  {
    List<File> list = new ArrayList<File>();
    
    searchFiles(file_name, file, 0, DEFAULT_MAX_DEPTH, list);
    
    return list;
  }
  
  /**
   * Mthode pour obtenir une liste d'accs  des rpertoire correspondant  un nom particulier.
   * 
   * @param directory_name Le nom du rpertoire en recherche.
   * @return Une liste de rpertoires ayant le nom dsir.
   */
  public List<File> searchDirectories(String directory_name)
  {
    List<File> list = new ArrayList<File>();
    
    searchDirectories(directory_name, this.file, 0, this.DEFAULT_MAX_DEPTH, list);
    
    return list;
  }
  
  /**
   * Mthode pour obtenir un accs  un fichier correspondant  un nom particulier  partir de l'emplacemenet de l'explorateur et des sous-rpertoires.
   * 
   * @param file_name Le nom du fichier en recherche (avec extension).
   * @return Un accs  un fichier ayant le nom dsir.
   * @throws FileNotFoundException Si le fichier n'a pas t trouv.
   * @throws SManyFilesFoundException Si plusieurs fichiers ayant le mme nom a t trouvs.
   */
  public File findFile(String file_name) throws FileNotFoundException, SManyFilesFoundException
  {
    List<File> list = searchFiles(file_name);
    
    if(list.isEmpty())
      throw new FileNotFoundException("Erreur SBrowser 003 : Le fichier '" + file_name + "' n'a pas t trouv  partir de l'emplacement initiale '" + getPath() + "'.");
    
    if(list.size() > 1)
      throw new SManyFilesFoundException("Erreur SBrowser 004 : Le fichier '" + file_name + "' a t trouv  pluseurs reprise  partir de l'emplacement initiale '" + getPath() + "'.");
    
    return list.get(0);
  }
  
  /**
   * Mthode pour localiser un rpertoire  partir de l'emplacement du browser. 
   * La recherche sera effectue rcursivement dans l'ensemble de l'arborsance des rpertoires.
   * Si le nom du rpertoire est localis  plusieurs endroits, une exception sera lance.
   * Si le nom du rpertoire n'est pas localis, une exception sera lance.
   * 
   * @param directory_name Le nom du rpertoire.
   * @return L'accs au rpertoire.
   * @throws FileNotFoundException Si le rpertoire n'a pas t trouv.
   * @throws SManyFilesFoundException Si le rpertoire a t trouv  plusieurs endroits.
   */
  public File findDirectory(String directory_name) throws FileNotFoundException, SManyFilesFoundException
  {
    List<File> list = searchDirectories(directory_name);
    
    if(list.isEmpty())
      throw new FileNotFoundException("Erreur SBrowser 004 : Le rpertoire '" + directory_name + "' n'a pas t trouv  partir de l'emplacement initiale '" + getPath() + "'.");
    
    if(list.size() > 1)
      throw new SManyFilesFoundException("Erreur SBrowser 005 : Le rpertoire '" + directory_name + "' a t trouv  pluseurs reprise  partir de l'emplacement initiale '" + getPath() + "'.");
    
    return list.get(0);
  }
  
  /**
   * <p>
   * Mthode pour dplacer une fichier dans un autre rpertoire.
   * Le nom du fichier sera cherch  partir de l'emplacement du browser
   * et sera dplac dans le rpertoire dsir  partir de l'emplacement du browser. 
   * </p>
   * <p>
   * Puisque cette mthode effectue une recherche du fichier <u>sans duplicata du nom admissible</u>,
   * il n'y a pas de risque que le fichier dplac  un autre endroit remplace un autre fichier existant ayant le mme nom.
   * </p>
   * 
   * @param file_name Le nom du fichier  dplacer.
   * @param directory_name Le nom du rpertoire o sera dplac le fichier.
   * @throws FileNotFoundException Si le fichier ou le rpertoire n'a pas t trouv.
   * @throws SManyFilesFoundException Si le nom du fichier ou rpertoire a t trouv  plusieurs endroits.
   */
  public void moveFile(String file_name, String directory_name) throws FileNotFoundException, SManyFilesFoundException
  {
    // Rechercher le fichier  dplacer.
    File file = findFile(file_name);
    
    // Obtenir la localisation du rpertoire vers o le fichier sera dplac. 
    File directory = findDirectory(directory_name);
    
    // Copier le fichier et effacer s'il y a eu copie.
    if(file.renameTo(new File(directory.getPath() + SStringUtil.FILE_SEPARATOR_CARACTER + file_name)))
      file.delete(); 
  }
    
  /**
   * Mthode pour crire un fichier texte  l'emplacement du browser.
   * 
   * @param file_name Le nom du fichier. 
   * @param extension L'extension du fichier.
   * @param text Le texte  crire dans le fichier.
   * @throws IOException S'il y a eu des erreur de type I/O.
   */
  public void writeTextFile(String file_name, String extension, String text) throws IOException
  {
    FileWriter fw = new FileWriter(this.createFile(file_name, extension));
    BufferedWriter bw = new BufferedWriter(fw);
    
    bw.write(text);
    bw.flush();
    
    bw.close();     
    fw.close();
  }
  
  /**
   * Mthode permettant d'crire un BufferedImage en fichier image  l'emplacement du browser.
   *
   * @param file_name Le nom du fichier (sans l'extension).
   * @param extension Le nom de l'extension du fichier (format de l'image gnr). 
   * @param image L'image.
   * @throws IOException Si l'image n'a pas pu tre convertie en fichier.
   */
  public void writeImageFile(String file_name, String extension, BufferedImage image) throws IOException
  {
    // Gnrer le fichier image
    File image_file = this.createFile(file_name, extension);
     
    // criture de l'image.
    ImageIO.write(image, extension, image_file);
  }
  
  /**
   * Mthode permettant de lire un fichier image afin d'en gnrer un BufferedImage.
   * 
   * @param file_name Le nom du fichier (avec extension).
   * @return L'image contenu dans le fichier.
   * @throws FileNotFoundException
   * @throws SManyFilesFoundException
   * @throws IOException
   */
  public BufferedImage readImageFile(String file_name) throws FileNotFoundException, SManyFilesFoundException, IOException
  {
    File file = this.findFile(file_name);
      
    return ImageIO.read(file);
  }
  
  /**
   * Mthode pour faire la recherche d'un fichier dsir de faon rcursive.  chaque fois que le fichier sera trouv,
   * l'adresse du fichier sera sauvegard dans la liste.
   * 
   * @param file_name Le nom du fichier en recherche (avec extension).
   * @param current_file Le rpertoire o la recherche est rendue.
   * @param depth Le niveau de profondeur courant de la rcursivit.
   * @param max_depth Le niveau de profoncteur maximal de la rcursivit.
   * @param file_list La liste des fichies trouvs.
   */
  private void searchFiles(final String file_name, final File current_file, final int depth, final int max_depth, final List<File> file_list)
  {
    // Vrification du niveau de profondeur maximal du fichier en recherche
    if(depth <= max_depth)
    {
      // Vrification si l'analyse se fait sur un rpertoire et s'il est accessible.
      if(current_file.isDirectory() && current_file.canRead())                            
      {
        // Itrer sur l'ensemble du contenu du rpertoire
        for(File temp : current_file.listFiles())                                 
        {
          // Vrifier si le fichier itr est un rpertoire. Dans ce cas, on continue la recherche rcursivement.  
          if(temp.isDirectory())                                         
            searchFiles(file_name, temp, depth + 1, max_depth, file_list); 
          else 
            if(temp.getName().equalsIgnoreCase(file_name))                // Si le fichier est trouv ! 
                file_list.add(temp);    
        }
      } 
    }
  }
  
  /**
   * Mthode pour faire la recherche du rpertoire dsir de faon rcursive.  chaque fois que le rpertoire sera trouv,
   * l'adresse du rpertoire sera sauvegard dans la liste.
   * 
   * @param file Le rpertoire o la recherche est rendue.
   * @param depth Le niveau de profondeur de la rcursivit.
   */
  private void searchDirectories(final String directory_name, final File current_file, final int depth, final int max_depth, final List<File> directory_list)
  {
    // Vrification du niveau de profondeur maximal du fichier en recherche
    if(depth <= max_depth)
    {
      // Vrification si l'analyse se fait sur un rpertoire et l'accs au rpertoire est possible.
      if(current_file.isDirectory() && current_file.canRead())                                                  
      {                                            
        // Vrifier si le rpertoire courant porte le nom recherch.
        if(directory_name.equalsIgnoreCase(current_file.getName()))
          directory_list.add(current_file);
        
        // Itrer sur l'ensemble du contenu du rpertoire et itrer rcursivement dans le cas d'un rpertoire.
        for(File temp : current_file.listFiles())   
          if(temp.isDirectory())            
            searchDirectories(directory_name, temp, depth + 1, max_depth, directory_list);        
      }
    }
  }
  
  /**
   * Mthode pour effacer recursivement un rpertoire et son contenu.
   * 
   * @param directory Le rpertoire  effacer et son contenu.
   */
  private void deleteDirectory(File directory)
  {
    for(File f : directory.listFiles())
      if(f.isDirectory())
        deleteDirectory(f);
      else
        f.delete();
    
    directory.delete();
  }
  
}
