/**
 * 
 */
package sim.graphics;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.FutureTask;
import java.util.concurrent.ThreadPoolExecutor;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.graphics.shader.SShader;
import sim.graphics.SAbstractRaytracer;
import sim.math.SVectorPixel;
import sim.util.SLog;

/**
 * La classe <b>SThreadPoolCastRaytracer</b> reprsentant un raytracer lanant <b>plusieurs rayons</b> dans chaque pixel du viewport. 
 * Cette mthode permet de rduire l'aliassage dans l'image puisque qu'une moyenne de la couleur 
 * obtenue par chaque rayon lanc dans chaque pixel sera utilis pour dfinir la couleur de chaque pixel.
 * 
 * @author Simon Vzina et Robin Vzina
 * @since 2015-04-07
 * @version 2016-01-13
 */
public class SThreadPoolCastRaytracer extends SAbstractRaytracer {

  private static final int DEFAULT_TASK = 2;			  //nombre de tches par exccution par dfaut (typiquement le nombre de processeur que l'on veut exploiter)
  private static final int DEFAULT_SAMPLING = 1;		//nombre de rayon lanc par pixel par dfaut
  
  private final int task;                           //nombre de tches par excution (nombre de thread en excution simultane)
  private final int sampling;	                      //nombre de rayon lanc dans chaque pixel
  
  private final ThreadPoolExecutor executor;        //excuteur des tches de calcul de la couleur de chaque pixel 
  
  /**
   * Sous-classe reprsentant une tche  raliser par un thread.
   * La tche  raliser sera de calculer une SColor et prendra comme paramtre un SVectorPixel.
   * @author Simon Vzina et Robin Vzina
   * @since 2015-04-07
   * @version 2015-04-10
   */
  protected final class FutureSColor implements Callable<SColor>
  {
    private SVectorPixel pixel;	//le pixel dont la couleur doit tre valu

    /**
     * Constructeur de la futur tche  raliser.
     * @param pixel - La coordonne du pixel dont la couleur sera  valuer.
     */
    public FutureSColor(SVectorPixel pixel) 
    {
      this.pixel = pixel;
    }

    @Override
    public SColor call() throws Exception
    {
      //La tche a raliser sera d'valuer la couleur du pixel (mthode protected de la classe prsente).
      return evaluatePixelColor(pixel);
    }
  }

  
  /**
   * Constructeur d'un raytracer de type <i>multi cast</i> avec <b>un rayon</b> par dfaut. 
   * 
   * @param view_frustum - La pyramide de vue.
   * @param shader - Le calculateur d'illumination (<i>shader</i>).
   * @param viewport - La fentre de rendu (<i>viewport</i>).
   */
  public SThreadPoolCastRaytracer(SViewFrustum view_frustum, SShader shader, SViewport viewport)
  {
    this(view_frustum, shader, viewport, DEFAULT_SAMPLING, DEFAULT_TASK);
  }

  /**
   * Constructeur d'un raytracer de type <i>multi cast</i>.
   * 
   * @param view_frustum - La pyramide de vue.
   * @param shader - Le calculateur d'illumination (<i>shader</i>).
   * @param viewport - La fentre de rendu (<i>viewport</i>).
   * @param sampling - Le nombre de rayon lanc par pixel afin de rduire le crnelage (<i>anti aliasing>/i>). 
   * @param task - Le nombre de tches simultans en calcul (pour le <i>multithreading</i>).
   * @throws SConstructorException Si le nombre de rayons par pixel (<i>sampling</i>) n'est pas suprieur  0.
   * @throws SConstructorException Si le nombre de tches (<i>task</i>) n'est pas suprieur  0.
   */
  public SThreadPoolCastRaytracer(SViewFrustum view_frustum, SShader shader, SViewport viewport, int sampling, int task)throws SConstructorException
  {
    super(view_frustum, shader, viewport);
    
    //Vrification que le sampling soit suprieur  0
    if(sampling > 0)
    	this.sampling = sampling;
    else
    	throw new SConstructorException("Erreur SMultiCastRaytracer 001 : Le nombre de rayon par pixel '" + sampling + "' n'est pas suprieur  0.");
    
    //Vrification que le sampling soit suprieur  0
    if(task > 0)
      this.task = task;
    else
      throw new SConstructorException("Erreur SMultiCastRaytracer 002 : Le nombre de rayon par pixel '" + task + "' n'est pas suprieur  0.");
      
    //Message de mauvaise fonctionnalit si la coordone interne d'un pixel n'est pas alatoire lors de super sampling
    if(sampling > 1 && view_frustum.getPixelInternalCoordinate() != SViewFrustum.RANDOM_PIXEL)
      SLog.logWriteLine("Message SMultiCastRaytracer : Puisque le sampling est  '" + sampling + "' et que la coordonne interne d'un pixel est '" + SViewFrustum.PIXEL_COORDINATE[view_frustum.getPixelInternalCoordinate()] + "'(code = " + view_frustum.getPixelInternalCoordinate() + ") et non pas '" + SViewFrustum.PIXEL_COORDINATE[SViewFrustum.RANDOM_PIXEL] + "'(code = " + SViewFrustum.RANDOM_PIXEL + "), l'antialiasing ne peut pas tre applicable.");
    
    //Construction de l'excuteur du threadPool
    executor = (ThreadPoolExecutor) Executors.newFixedThreadPool(task);
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see sim.graphics.SRaytracer#raytrace(int)
   */
  @Override
  public void raytrace(int nb_pixel) throws SRuntimeException
  {
    //Vrification du nombre de pixels demand en calcul
    if(nb_pixel < 0)
      throw new SRuntimeException("Erreur SMultiCastRaytracer 003 : Le nombre de pixels '" + nb_pixel + "' doit tre suprieur  0.");

    int count = 0; // compter le nombre de pixel calcul

    //Excuter des calculs tant que le nombre de pixel dsir n'est pas atteint
    while(count < nb_pixel)
    {
      // S'il y a encore des pixels disponibles  tre dessin
      if (viewport.hasNextPixel()) 
      {
        // le prochain pixel  calculer
        List<SVectorPixel> listSVectorPixel = new ArrayList<SVectorPixel>();
        for(int i = 0; i < task && viewport.hasNextPixel(); i++)
        {
          SVectorPixel nextPixel = viewport.nextPixel();
          listSVectorPixel.add(nextPixel);
        }

        
        List<FutureTask<SColor>> list = new ArrayList<FutureTask<SColor>>();
        for(SVectorPixel sVectorPixel : listSVectorPixel)
          list.add(new FutureTask<SColor>(new FutureSColor(sVectorPixel)));

        //Chargement des tches  faire dans l'excuteur et lancement de la thread
        for (FutureTask<SColor> f : list)
          executor.execute(f);

        List<SColor> listSColor = new ArrayList<SColor>();

        // Dterminer la couleur calcule par le shader pour le pixel
        try{
          
          for(FutureTask<SColor> f : list)
            listSColor.add(f.get());
            
        } catch (InterruptedException e) {
          throw new SRuntimeException("Erreur SThreadPoolCastRaytracer 004 : Raytracer interropu.", e);
        } catch (ExecutionException e) {
          throw new SRuntimeException("Erreur SThreadPoolCastRaytracer 005 : Erreur d'excution.", e);
        }

        // Dessiner les pixels calculs dans le viewport
        for(int i = 0; i < listSColor.size(); i++) 
          viewport.setColor(listSVectorPixel.get(i), listSColor.get(i));
        
        count += listSColor.size(); // augmenter le compteur
      } 
      else                         
        count = nb_pixel;           // nombre de pixel  calcul atteint, donc on arrte tout
    }
  }

  @Override
  protected SColor evaluatePixelColor(SVectorPixel pixel)
  {
    //--------------------------------------------------------------------
    //valuer la couleur d'un pixel selon le nombre de sampling  valuer
    //--------------------------------------------------------------------
    
    if(sampling == 1)                         // un rayon par pixel
      return raytracePixel(pixel);
    else                                      // plusieurs rayons par pixel
    {
      SColor[] tab = new SColor[sampling];
      
      //Calcul la couleur pour chaque sampling
      for(int i=0; i<tab.length; i++)
      	tab[i] = raytracePixel(pixel);
      
      //Faire la somme des couleurs
      SColor sum = new SColor(0.0, 0.0, 0.0);
      
      for(int i=0; i<tab.length; i++)
      	sum = sum.add(tab[i]);
      
      //Retourner la couleur aprs avoir divis par le nombre de sampling
      return sum.multiply( 1.0 / (double)sampling);
    }
  }
  
}// fin classe SSingleCastRaytracer
